package edu.stanford.nlp.fsm;

import edu.stanford.nlp.fsm.TransducerGraph;
import edu.stanford.nlp.util.Pair;
import edu.stanford.nlp.util.Maps;
import edu.stanford.nlp.util.Timing;

import java.util.*;

/**
 * Minimization in n log n a la Hopcroft.
 *
 * @author Dan Klein (klein@cs.stanford.edu)
 */
public class ExactAutomatonMinimizer implements AutomatonMinimizer {
  TransducerGraph unminimizedFA = null;
  Map memberToBlock = null;
  LinkedList activePairs = null;


  boolean sparseMode = false;

  static final Object SINK_NODE = new Object();

  protected TransducerGraph getUnminimizedFA() {
    return unminimizedFA;
  }

  protected Collection getSymbols() {
    return getUnminimizedFA().getInputs();
  }

  public TransducerGraph minimizeFA(TransducerGraph unminimizedFA) {
    this.unminimizedFA = unminimizedFA;
    this.activePairs = new LinkedList();
    this.memberToBlock = new HashMap();
    minimize();
    return buildMinimizedFA();
  }

  protected TransducerGraph buildMinimizedFA() {
    TransducerGraph minimizedFA = new TransducerGraph();
    TransducerGraph unminimizedFA = getUnminimizedFA();
    for (Iterator arcI = unminimizedFA.getArcs().iterator(); arcI.hasNext();) {
      TransducerGraph.Arc arc = (TransducerGraph.Arc) arcI.next();
      Object source = projectNode(arc.getSourceNode());
      Object target = projectNode(arc.getTargetNode());
      try {
        if (minimizedFA.canAddArc(source, target, arc.getInput(), arc.getOutput()))
          minimizedFA.addArc(source, target, arc.getInput(), arc.getOutput());
      } catch(Exception e) {
        //throw new IllegalArgumentException();
      }
    }
    minimizedFA.setStartNode(projectNode(unminimizedFA.getStartNode()));
    for (Iterator endIter = unminimizedFA.getEndNodes().iterator(); endIter.hasNext();) {
      Object o = (Object) endIter.next();
      minimizedFA.setEndNode(projectNode(o));
    }

    return minimizedFA;
  }

  protected Object projectNode(Object node) {
    return getBlock(node).getMembers();
  }



  protected boolean hasActivePair() {
    return activePairs.size() > 0;
  }

  protected Pair getActivePair() {
    return (Pair) activePairs.removeFirst();
  }

  protected void addActivePair(Pair pair) {
    activePairs.addLast(pair);
  }

//  protected Collection inverseImages(Collection block, Object symbol) {
//    List inverseImages = new ArrayList();
//    for (Iterator nodeI = block.iterator(); nodeI.hasNext();) {
//      Object node = nodeI.next();
//      inverseImages.addAll(getUnminimizedFA().getInboundArcs(node, symbol));
//    }
//    return inverseImages;
//  }

  protected Map sortIntoBlocks(Collection nodes) {
    Map blockToMembers = new HashMap(); // IdentityHashMap();
    for (Iterator nodeI = nodes.iterator(); nodeI.hasNext();) {
      Object o = nodeI.next();
      ExactBlock block = getBlock(o);
      if (block==null)
	throw new RuntimeException("got null block");
      Maps.putIntoValueHashSet(blockToMembers, block, o);
    }
    return blockToMembers;
  }

  protected void makeBlock(Collection members) {
    ExactBlock block = new ExactBlock(new HashSet(members));
    for (Iterator memberI = block.getMembers().iterator(); memberI.hasNext();) {
      Object member = memberI.next();
      if (member != SINK_NODE)
        memberToBlock.put(member, block);
    }
    for (Iterator symbolI = getSymbols().iterator(); symbolI.hasNext();) {
      Object symbol = symbolI.next();
      addActivePair(new Pair(block, symbol));
    }
  }

  protected void removeAll(Collection block, Collection members) {
    // this is because AbstractCollection/Set.removeAll() isn't always linear in members.size()
    for (Iterator memberI = members.iterator(); memberI.hasNext();) {
      Object member = memberI.next();
      block.remove(member);
    }
  }

  protected Collection difference(Collection block, Collection members) {
    Set difference = new HashSet();
    for (Iterator memberI = block.iterator(); memberI.hasNext();) {
      Object member = memberI.next();
      if (! members.contains(member))
        difference.add(member);
    }
    return difference;
  }

  protected ExactBlock getBlock(Object o) {
    ExactBlock result = (ExactBlock) memberToBlock.get(o);
    if (result == null)
      throw new RuntimeException("memberToBlock had null block");
    return result;
  }

  protected Collection getInverseImages(ExactBlock block, Object symbol) {
    List inverseImages = new ArrayList();
    for (Iterator memberI = block.getMembers().iterator(); memberI.hasNext();) {
      Object member = memberI.next();
      Collection arcs = null;
      if (member != SINK_NODE) {
        arcs = getUnminimizedFA().getArcsByTargetAndInput(member, symbol);
      } else {
        arcs = getUnminimizedFA().getArcsByInput(symbol);
        if (! sparseMode)
          arcs = difference(getUnminimizedFA().getArcs(), arcs);
      }
      if (arcs == null)
        continue;
      for (Iterator arcI = arcs.iterator(); arcI.hasNext();) {
        TransducerGraph.Arc arc = (TransducerGraph.Arc) arcI.next();
        Object source = arc.getSourceNode();
        inverseImages.add(source);
      }
    }
    return inverseImages;
  }

  protected void makeInitialBlocks() {
    // sink block (for if the automaton isn't complete
    makeBlock(Collections.singleton(SINK_NODE));
    // accepting block
    Set endNodes = getUnminimizedFA().getEndNodes();
    makeBlock(endNodes);
    // main block
    Collection nonFinalNodes = new HashSet(getUnminimizedFA().getNodes());
    nonFinalNodes.removeAll(endNodes);
    makeBlock(nonFinalNodes);
  }

  protected void minimize() {
    makeInitialBlocks();
    while (hasActivePair()) {
      Pair activePair = getActivePair();
      ExactBlock activeBlock = (ExactBlock) activePair.first();
      Object symbol = activePair.second();
      Collection inverseImages = getInverseImages(activeBlock, symbol);
      Map inverseImagesByBlock = sortIntoBlocks(inverseImages);
      for (Iterator blockI = inverseImagesByBlock.keySet().iterator(); blockI.hasNext();) {
        ExactBlock block = (ExactBlock) blockI.next();
	if (block==null)
	  throw new RuntimeException("block was null");
        Collection members = (Collection) inverseImagesByBlock.get(block);
        if (members.size() == 0 || members.size() == block.getMembers().size())
          continue;
        if (members.size() > block.getMembers().size() - members.size()) {
          members = difference(block.getMembers(), members);
        }
        removeAll(block.getMembers(), members);
        makeBlock(members);
      }
    }
  }

  public ExactAutomatonMinimizer(boolean sparseMode) {
    this.sparseMode = sparseMode;
  }

  public ExactAutomatonMinimizer() {
    this(false);
  }

  public static void main(String[] args) {
    TransducerGraph fa = new TransducerGraph();
    fa.addArc(fa.getStartNode(),"1","a","");
    fa.addArc(fa.getStartNode(),"2","b","");
    fa.addArc(fa.getStartNode(),"3","c","");
    fa.addArc("1","4","a","");
    fa.addArc("2","4","a","");
    fa.addArc("3","5","c","");
    fa.addArc("4","6","c","");
    fa.addArc("5","6","c","");
    fa.setEndNode("6");
    System.out.println(fa);
    ExactAutomatonMinimizer minimizer = new ExactAutomatonMinimizer();
    System.out.println(minimizer.minimizeFA(fa));
    System.out.println("Starting...");
    Timing.startTime();
    TransducerGraph randomFA = TransducerGraph.createRandomGraph(100, 10, 1.0, 10, new ArrayList());
    TransducerGraph minimizedRandomFA = minimizer.minimizeFA(randomFA);
    System.out.println(randomFA);
    System.out.println(minimizedRandomFA);
    Timing.tick("done. ( "+randomFA.getArcs().size()+" arcs to "+minimizedRandomFA.getArcs().size()+" arcs)");
  }
}


class ExactBlock implements Block {
    Set members;
    public Set getMembers() {
      if (members == null)
	throw new RuntimeException("tried to create block with null members.");
      return members;
    }
    public ExactBlock(Set members) {
      if (members == null)
	throw new RuntimeException("tried to create block with null members.");
      this.members = members;
    }
    public String toString() {
      return "Block: " + members.toString();
    }
  }

