package edu.stanford.nlp.parser.lexparser;

import edu.stanford.nlp.trees.*;
import java.io.*;
import java.util.*;


/** An abstract class providing a common method base from which to
 * complete a <code>TreebankLangParserParams</code> implementing class.
 * <p>
 *
 * With some extending classes you'll want to have access to special
 * attributes of the corresponding TreebankLanguagePack while taking
 * advantage of this class's code for making the TreebankLanguagePack
 * accessible.  A good way to do this is to pass a new instance of the
 * appropriate TreebankLanguagePack into this class's constructor,
 * then get it back later on by casting a call to
 * treebankLanguagePack().  See ChineseTreebankParserParams for an
 * example.
 *
 * @author Roger Levy
 */
public abstract class AbstractTreebankParserParams implements TreebankLangParserParams {
  
  protected String inputEncoding;
  protected String outputEncoding;
  protected TreebankLanguagePack tlp;

  /** Stores the passed-in TreebankLanguagePack. */
  protected AbstractTreebankParserParams(TreebankLanguagePack tlp) {
    this.tlp = tlp;
  }

  /** sets the input encoding */
  public void setInputEncoding(String encoding) {
    inputEncoding = encoding;
  }

  /** sets the output encoding */
  public void setOutputEncoding(String encoding) {
    outputEncoding = encoding;
  }



  /** returns a MemoryTreebank appropriate to the treebank source */
  public abstract MemoryTreebank memoryTreebank();


  /** You can often return the same thing for testMemoryTreebank as
   * for memoryTreebank */
  public MemoryTreebank testMemoryTreebank() {
    return memoryTreebank();
  }


  /** the PrintWriter used to print output. It's the responsibility of
   * pw to deal properly with character encodings for the relevant
   * treebank */
  public PrintWriter pw() {
    return pw(System.out);
  }

  /** the PrintWriter used to print output. It's the responsibility of
   * pw to deal properly with character encodings for the relevant
   * treebank */
  public PrintWriter pw(OutputStream o) {
    String encoding = outputEncoding;
    if(! java.nio.charset.Charset.isSupported(encoding)) {
      System.out.println("Warning: desired encoding " + encoding + " not accepted. ");
      System.out.println("Using UTF-8 to construct PrintWriter");
      encoding = "UTF-8";
    }
      
    try{
      return new PrintWriter(new OutputStreamWriter(o,
						    encoding), true);
    } catch(UnsupportedEncodingException e) {
      System.out.println("Warning: desired encoding " + outputEncoding + " not accepted. "  +  e);
      try{
	return new PrintWriter(new OutputStreamWriter(o,"UTF-8"),true);
      }
      catch(UnsupportedEncodingException e1) {
	System.out.println("Something is really wrong.  Your system doesn't even support UTF-8!" + e1);
	return new PrintWriter(o,true);
      }
    }
  }


  /** Returns an appropriate treebankLanguagePack */
  public TreebankLanguagePack treebankLanguagePack() {
    return tlp;
  }


  /** the HeadFinder to use for your treebank. */
  public abstract HeadFinder headFinder();

  public abstract Lexicon lex();
 
  /** the tree transformer used to produce trees for evaluation.  Will
   * be applied both to the parse output tree and to the gold
   * tree. Should strip punctuation and maybe do some other things. */

  public abstract TreeTransformer collinizer();

  /** the tree transformer used to produce trees for evaluation.  Will
   * be applied both to the parse output tree and to the gold
   * tree. Should strip punctuation and maybe do some other
   * things. The evalb version should strip some more stuff
   * off. (finish this doc!) */
  public abstract TreeTransformer collinizerEvalb();

  
  

  /** Returns the splitting strings used for selective splits.
   *  @return An array containing ancestor-annotated Strings: categories
   *    should be split according to these ancestor annotations.
   */
  public abstract String[] splitters();

  /** Returns the splitting strings used for selective splits.
   *  @return An array containing ancestor-annotated Strings: categories
   *    should be split according to these ancestor annotations.
   */
  public abstract String[] sisterSplitters();


  /** transformTree does language-specific tree transformations such
   *  as splicing. Any parameterizations should be inside the specific
   * TreebankLangParserParams class 
   */
  public abstract TreeHeadPair transformTree(Tree t, Tree root, TreeHeadPair thp);

  /** display language-specific settings */
  public abstract void display();

  /** Set language-specific options according to flags.
   *  This routine should process the option starting in args[i] (which
   *  might potentially be several arguments long if it takes arguments).
   *  It should return the index after the last index it consumed in
   *  processing.  In particular, if it cannot process the current option,
   *  the return value should be i.
   */
  public abstract int setOptionFlag(String[] args, int i);

}
