package edu.stanford.nlp.parser.lexparser;

import edu.stanford.nlp.util.Numberer;

import java.io.*;
import java.util.*;

/** Maintains efficient indexing of binary grammar rules.
 * 
 *  @author Dan Klein
 */
public class BinaryGrammar implements Serializable {

  private String stateSpace;

  private int numStates;
  private transient List[] rulesWithParent = null;
  private transient List[] rulesWithLC = null;
  private transient List[] rulesWithRC = null;
  private List allRules = null;
  private transient BinaryRule[][] splitRulesWithLC = null;
  private transient BinaryRule[][] splitRulesWithRC = null;
  private transient BinaryRule[][] splitRulesWithParent = null;
  private transient Map ruleMap = null;


  public int numRules() { return allRules.size(); }
  public List rules() { return allRules; }

  public String stateSpace() {
    return stateSpace;
  }

  // for super speed! (maybe)
  private Numberer stateNumberer;
  private boolean[] synthetic = null;

  public boolean isSynthetic(int state) {
    if (synthetic == null) {
      synthetic = new boolean[numStates];
      for (int s = 0; s < numStates; s++) {
        try {
          //System.out.println(((String)stateNumberer.object(s))); // debugging
          synthetic[s] = (((String) stateNumberer.object(s)).charAt(0) == '@');
        } catch (NullPointerException e) {
          synthetic[s] = true;
        }
      }
    }
    return synthetic[state];
  }

  private BinaryRule[] toBRArray(List list) {
    BinaryRule[] array = new BinaryRule[list.size()];
    for (int i = 0; i < array.length; i++)
      array[i] = (BinaryRule) list.get(i);
    return array;
  }

  public void splitRules() {
    splitRulesWithLC = new BinaryRule[numStates][];
    splitRulesWithRC = new BinaryRule[numStates][];
    splitRulesWithParent = new BinaryRule[numStates][];
    // rules accesed by their "synthetic" child or left child if none
    for (int state = 0; state < numStates; state++) {
      //      System.out.println("Splitting rules for state: " + stateNumberer.object(state));
      // check synthetic
      if (isSynthetic(state)) {
        splitRulesWithLC[state] = toBRArray(rulesWithLC[state]);
        splitRulesWithRC[state] = toBRArray(rulesWithRC[state]);
      } else {
        List ruleList = new ArrayList();
        for (int rNum = 0; rNum < rulesWithLC[state].size(); rNum++) {
          BinaryRule br = (BinaryRule) rulesWithLC[state].get(rNum);
          if (!isSynthetic(br.rightChild))
            ruleList.add(br);
        }
        splitRulesWithLC[state] = toBRArray(ruleList);
        ruleList = new ArrayList();
        for (int rNum = 0; rNum < rulesWithRC[state].size(); rNum++) {
          BinaryRule br = (BinaryRule) rulesWithRC[state].get(rNum);
          if (!isSynthetic(br.leftChild))
            ruleList.add(br);
        }
        splitRulesWithRC[state] = toBRArray(ruleList);
      }
      // parent accessor
      splitRulesWithParent[state] = toBRArray(rulesWithParent[state]);
    }
  }

  public BinaryRule[] splitRulesWithLC(int state) {
    return splitRulesWithLC[state];
  }

  public BinaryRule[] splitRulesWithRC(int state) {
    return splitRulesWithRC[state];
  }

  public BinaryRule[] splitRulesWithParent(int state) {
    return splitRulesWithParent[state];
  }

  // the sensible version

  public double scoreRule(BinaryRule br) {
    BinaryRule rule = (BinaryRule) ruleMap.get(br);
    return (rule != null ? rule.score : Double.NEGATIVE_INFINITY);
  }

  public void addRule(BinaryRule br) {
    //    System.out.println("BG adding rule " + br);
    rulesWithParent[br.parent].add(br);
    rulesWithLC[br.leftChild].add(br);
    rulesWithRC[br.rightChild].add(br);
    allRules.add(br);
    ruleMap.put(br, br);
  }


  public Iterator ruleIterator() {
    return allRules.iterator();
  }

  public Iterator ruleIteratorByParent(int state) {
    return rulesWithParent[state].iterator();
  }

  public Iterator ruleIteratorByRightChild(int state) {
    return rulesWithRC[state].iterator();
  }

  public Iterator ruleIteratorByLeftChild(int state) {
    return rulesWithLC[state].iterator();
  }

  public List ruleListByParent(int state) {
    return rulesWithParent[state];
  }

  public List ruleListByRightChild(int state) {
    return rulesWithRC[state];
  }

  public List ruleListByLeftChild(int state) {
    return rulesWithLC[state];
  }

  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject();
    init();
    for (int i = 0; i < allRules.size(); i++) {
      BinaryRule br = (BinaryRule) allRules.get(i);
      rulesWithParent[br.parent].add(br);
      rulesWithLC[br.leftChild].add(br);
      rulesWithRC[br.rightChild].add(br);
      ruleMap.put(br, br);
    }
    splitRules();
  }

  void init() {
    stateNumberer = Numberer.getGlobalNumberer(stateSpace);
    ruleMap = new HashMap();
    rulesWithParent = new List[numStates];
    rulesWithLC = new List[numStates];
    rulesWithRC = new List[numStates];
    for (int s = 0; s < numStates; s++) {
      rulesWithParent[s] = new ArrayList();
      rulesWithLC[s] = new ArrayList();
      rulesWithRC[s] = new ArrayList();
    }
  }

  public BinaryGrammar(int numStates) {
    this(numStates, "states");
  }

  public BinaryGrammar(int numStates, String stateSpace) {
    this.stateSpace = stateSpace;
    this.numStates = numStates;
    allRules = new ArrayList();
    init();
  }

  /**
   * Populates data in this BinaryGrammar from the character stream
   * given by the Reader r.
   */
  public void readData(BufferedReader in) throws IOException {
    String line, parentString, leftChildString, rightChildString, scoreString;
    int lineNum = 1;
    StringTokenizer tok;
    Numberer n = Numberer.getGlobalNumberer("states");
// all lines have one rule per line
    line = in.readLine();
    while (line != null && !line.startsWith("**********")) {
      try {
        tok = new StringTokenizer(line, "\t"); // break on tabs only
        parentString = tok.nextToken();
        tok.nextToken();
        leftChildString = tok.nextToken();
        rightChildString = tok.nextToken();
        scoreString = tok.nextToken();
        addRule(new BinaryRule(n.number(parentString), n.number(leftChildString), n.number(rightChildString),
                Double.parseDouble(scoreString)));
      } catch (Exception e) {
        throw new IOException("Error on line " + lineNum);
      }
      lineNum++;
      line = in.readLine();
    }
    splitRules();
  }

  /**
   * Writes out data from this Object to the Writer w.
   */
  public void writeData(Writer w) throws IOException {
    PrintWriter out = new PrintWriter(w);
// all lines have one rule per line
    for (Iterator rI = ruleIterator(); rI.hasNext();) {
      out.println(((BinaryRule) rI.next()).toString());
    }
    out.flush();
  }

  private static final long serialVersionUID = 1L;

} // end class BinaryGrammar

