package edu.stanford.nlp.parser.lexparser;

import edu.stanford.nlp.util.Numberer;
import edu.stanford.nlp.util.Counter;
import edu.stanford.nlp.trees.Tree;
import edu.stanford.nlp.trees.HasTag;
import edu.stanford.nlp.trees.HasWord;

import java.io.*;
import java.text.NumberFormat;
import java.util.*;

public class DependencyGrammar implements Serializable {

  private TagProjection tagProjection = null;

  private static Numberer tagNumberer = Numberer.getGlobalNumberer("tags");
  private static Numberer wordNumberer = Numberer.getGlobalNumberer("words");

  protected int numWords;
  protected int numTags;

  protected int numTagBins;
  protected int[] tagBin;

  transient private Counter argCounter;
  transient private Counter stopCounter;
  private Counter coreDependencies = new Counter();
  //private Counter rheadCounter;
  //private Counter lheadCounter;
  //private Counter argCounter;
  //private Counter headArgCounter;
  //private Counter headDistCounter;
  //private double total = 0.0;

  public double[] smooth = {16.0, 16.0, 4.0};
  public double[] interp = {0.6, 0.6};
  public double distanceDecay = 0.0;


  public String toString() {
    NumberFormat nf = NumberFormat.getNumberInstance();
    nf.setMaximumFractionDigits(2);
    StringBuffer sb = new StringBuffer(2000);
    String cl = getClass().getName();
    sb.append(cl.substring(cl.lastIndexOf(".") + 1) +
            "[tags=" + numTags + ",words=" + numWords + "; head -> arg\n");
    for (Iterator dI = coreDependencies.keySet().iterator(); dI.hasNext();) {
      Dependency d = (Dependency) dI.next();
      double count = coreDependencies.getCount(d);
      sb.append(d + " count " + nf.format(count));
      if (dI.hasNext()) {
        sb.append(",");
      }
      sb.append("\n");
    }
    sb.append("]");
    return sb.toString();
  }

  public int numTagBins() {
    return numTagBins;
  }

  public int tagBin(int tag) {
    if (tag < 0)
      return tag;
    return tagBin[tag];
  }

  public static boolean pruneTW(IntTaggedWord argTW) {
    String[] punctTags = Options.get().langpack().punctuationTags();
    for (int i = 0; i < punctTags.length; i++) {
      if (argTW.tag == tagNumberer.number(punctTags[i])) {
        return true;
      }
    }
    return false;
  }

  public static boolean rootTW(IntTaggedWord rTW) {
    return rTW.tag == tagNumberer.number(Lexicon.BOUNDARY_TAG);
  }

  static class EndHead {
    public int end;
    public int head;
  }

  protected static transient EndHead tempEndHead = new EndHead();

  protected static EndHead treeToDependencyHelper(Tree tree, List depList,
                                                  int loc) {
//       try {
// 	PrintWriter pw = new PrintWriter(new OutputStreamWriter(System.out,"GB18030"),true);
// 	tree.pennPrint(pw);
//       }
//       catch (UnsupportedEncodingException e) {}


    if (tree.isLeaf() || tree.isPreTerminal()) {
      tempEndHead.head = loc;
      tempEndHead.end = loc + 1;
      return tempEndHead;
    }
    if (tree.children().length == 1)
      return treeToDependencyHelper(tree.children()[0], depList, loc);
    int start = loc;
    tempEndHead = treeToDependencyHelper(tree.children()[0], depList, loc);
    int lHead = tempEndHead.head;
    int split = tempEndHead.end;
    tempEndHead = treeToDependencyHelper(tree.children()[1], depList, tempEndHead.end);
    int end = tempEndHead.end;
    int rHead = tempEndHead.head;
    String hTag = ((HasTag) tree.label()).tag();
    String lTag = ((HasTag) tree.children()[0].label()).tag();
    String rTag = ((HasTag) tree.children()[1].label()).tag();
    String hWord = ((HasWord) tree.label()).word();
    String lWord = ((HasWord) tree.children()[0].label()).word();
    String rWord = ((HasWord) tree.children()[1].label()).word();
    boolean leftHeaded = hWord.equals(lWord);
    String aTag = (leftHeaded ? rTag : lTag);
    String aWord = (leftHeaded ? rWord : lWord);
    int hT = tagNumberer.number(hTag);
    int aT = tagNumberer.number(aTag);
    int hW = (wordNumberer.hasSeen(hWord) ?
            wordNumberer.number(hWord) :
            wordNumberer.number(Lexicon.UNKNOWN_WORD));
    int aW = (wordNumberer.hasSeen(aWord) ?
            wordNumberer.number(aWord) :
            wordNumberer.number(Lexicon.UNKNOWN_WORD));
    int head = (leftHeaded ? lHead : rHead);
    int arg = (leftHeaded ? rHead : lHead);
    Dependency dependency = new Dependency(hW, hT, aW, aT, leftHeaded, (leftHeaded ? split - head - 1 : head - split));
    depList.add(dependency);
    Dependency stopL = new Dependency(aW, aT, -2, -2, false, (leftHeaded ? arg - split : arg - loc));
    depList.add(stopL);
    Dependency stopR = new Dependency(aW, aT, -2, -2, true, (leftHeaded ? end - arg - 1 : split - arg - 1));
    depList.add(stopR);
    //System.out.println("Adding: "+dependency+" at "+tree.label());
    tempEndHead.head = head;
    return tempEndHead;
  }

  public static List treeToDependencyList(Tree tree) {
    List depList = new ArrayList();
    treeToDependencyHelper(tree, depList, 0);
    if (false) {
      System.out.println("----------------------------");
      tree.pennPrint();
      System.out.println(depList);
    }
    return depList;
  }

  protected double scoreAll(Collection deps) {
    double totalScore = 0.0;
    for (Iterator i = deps.iterator(); i.hasNext();) {
      Dependency d = (Dependency) i.next();
      //if (d.head.word == Numberer.number("words","via") ||
      //          d.arg.word == Numberer.number("words","via"))
      //System.out.println(d+" at "+score(d));
      double score = score(d);
      if (score > Double.NEGATIVE_INFINITY)
        totalScore += score;
    }
    return totalScore;
  }

  public void tune(List deps) {
    double bestScore = Double.NEGATIVE_INFINITY;
    double[] bestSmooth = {0.0, 0.0, 0.0};
    double[] bestInterp = {0.0, 0.0};
    double bestDecay = 0.0;
    double maxDecay = (Options.get().distance ? -0.01 : 0.0);
    for (smooth[0] = 16; smooth[0] < 32; smooth[0] *= 2.0) {
      for (smooth[1] = 16; smooth[1] < 32; smooth[1] *= 2.0) {
        for (smooth[2] = 4; smooth[2] <= (Options.get().distance ? 5 : 5); smooth[2] *= 2.0) {
          for (interp[0] = 0.6; interp[0] < 0.7; interp[0] += 0.1) {
            for (distanceDecay = maxDecay; distanceDecay == maxDecay; distanceDecay -= 0.01) {
              //for (interp[1]=0.1; interp[1]<0.9; interp[1] += 0.1) {
              double score = scoreAll(deps);
              if (Test.verbose)
                System.out.println("Trying s0 " + smooth[0] + " s1 " + smooth[1] + " i0 " + interp[0] + " s2 " + smooth[2] + " score " + score);
              if (score > bestScore) {
                System.arraycopy(smooth, 0, bestSmooth, 0, smooth.length);
                System.arraycopy(interp, 0, bestInterp, 0, interp.length);
                bestDecay = distanceDecay;
                bestScore = score;
              }
              //}
            }
          }
        }
      }
    }
    System.arraycopy(bestSmooth, 0, smooth, 0, bestSmooth.length);
    System.arraycopy(bestInterp, 0, interp, 0, bestInterp.length);
    distanceDecay = bestDecay;
    if (Test.verbose)
      System.out.println("Tuning selected s0 " + smooth[0] + " s1 " + smooth[1] + " s2 " + smooth[2] + " i0 " + interp[0]);
  }

  private static final IntTaggedWord stopTW = new IntTaggedWord(-2, -2);
  private static final IntTaggedWord wildTW = new IntTaggedWord(-1, -1);

  public void addRule(Dependency dependency, double count) {
    if (!Options.get().directional)
      dependency.leftHeaded = false;
    coreDependencies.incrementCount(dependency, count);
    /*new Dependency(dependency.head.word,
                                        dependency.head.tag,
                                        dependency.arg.word,
                                        dependency.arg.tag,
                                        dependency.leftHeaded,
                                        dependency.distance), count);
    */
    expandDependency(dependency, count);
  }


  private transient Dependency tempDependency = null;
  private transient Map expandDependencyMap = new HashMap();

  private Dependency intern(IntTaggedWord headTW, IntTaggedWord argTW, boolean leftHeaded, int dist) {
    Map map = expandDependencyMap;
    if (tempDependency == null)
      tempDependency = new Dependency();
    tempDependency.head = headTW;
    tempDependency.arg = argTW;
    tempDependency.leftHeaded = leftHeaded;
    tempDependency.distance = (short) dist;
    Dependency returnDependency = tempDependency;
    if (map != null) {
      returnDependency = (Dependency) map.get(tempDependency);
      if (returnDependency == null)
        map.put(tempDependency, tempDependency);
      returnDependency = tempDependency;
    }
    if (returnDependency == tempDependency)
      tempDependency = null;
    return returnDependency;
  }

  private transient List tagITWList = null;//new ArrayList();

  private void expandDependency(Dependency dependency, double count) {
    //if (Test.prunePunc && pruneTW(dependency.arg))
    //  return;
    if (dependency.head == null)
      return;
    if (dependency.arg == null)
      return;
    if (tagITWList == null)
      tagITWList = new ArrayList();
    while (dependency.head.tag + 2 >= tagITWList.size() || dependency.arg.tag + 2 >= tagITWList.size())
      tagITWList.add(null);
    IntTaggedWord headT = (IntTaggedWord) tagITWList.get(dependency.head.tag + 2);
    if (headT == null) {
      headT = new IntTaggedWord(-1, tagBin(dependency.head.tag));
      tagITWList.set(dependency.head.tag + 2, headT);
    }
    IntTaggedWord argT = (IntTaggedWord) tagITWList.get(dependency.arg.tag + 2);
    if (argT == null) {
      argT = new IntTaggedWord(-1, tagBin(dependency.arg.tag));
      tagITWList.set(dependency.arg.tag + 2, argT);
    }
    IntTaggedWord head = new IntTaggedWord(dependency.head.word, tagBin(dependency.head.tag));//dependency.head;
    IntTaggedWord arg = new IntTaggedWord(dependency.arg.word, tagBin(dependency.arg.tag));//dependency.arg;
    boolean leftHeaded = dependency.leftHeaded;
    short valBinDist = valenceBin(dependency.distance);
    short distBinDist = distanceBin(dependency.distance);

    boolean isStop = (arg.word == -2);
    if (!isStop) {
      argCounter.incrementCount(intern(head, arg, leftHeaded, valBinDist), count);
      argCounter.incrementCount(intern(headT, arg, leftHeaded, valBinDist), count);
      argCounter.incrementCount(intern(head, argT, leftHeaded, valBinDist), count);
      argCounter.incrementCount(intern(headT, argT, leftHeaded, valBinDist), count);

      argCounter.incrementCount(intern(head, wildTW, leftHeaded, valBinDist), count);
      argCounter.incrementCount(intern(headT, wildTW, leftHeaded, valBinDist), count);

      argCounter.incrementCount(intern(wildTW, arg, false, -1), count);
      argCounter.incrementCount(intern(wildTW, argT, false, -1), count);
    }
    stopCounter.incrementCount(intern(head, arg, leftHeaded, distBinDist), count);
    stopCounter.incrementCount(intern(head, wildTW, leftHeaded, distBinDist), count);
    stopCounter.incrementCount(intern(headT, arg, leftHeaded, distBinDist), count);
    stopCounter.incrementCount(intern(headT, wildTW, leftHeaded, distBinDist), count);
  }

  protected static short valenceBin(int distance) {
    if (!Options.get().distance)
      return 0;
    if (distance < 0)
      return -1;
    if (distance == 0)
      return 0;
    return 1;
  }

  public static int NUMDISTBIN() {
    return Options.get().coarseDistance ? 4 : 5;
  }

  protected static short distanceBin(int distance) {
    if (!Options.get().distance)
      return 0;
    else if (Options.get().coarseDistance) {
      if (distance <= 0)
        return 0;
      if (distance <= 2)
        return 1;
      if (distance <= 5)
        return 2;
      else
        return 3;
    } else {
      if (distance <= 0)
        return 0;
      if (distance <= 1)
        return 1;
      if (distance <= 5)
        return 2;
      if (distance <= 10)
        return 3;
      else
        return 4;
    }
  }

  /*
  protected Dependency tempDependency = new Dependency(-2,-2,-2,-2,false,0);
  public double oldSscore(Dependency dependency) {
    if (! Test.directional)
      dependency.leftHeaded = false;
    // score the link, plus the stops for the arg
    double score = 0.0;
    score += rawScore(dependency);
    tempDependency.head.word = dependency.arg.word;
    tempDependency.head.tag = dependency.arg.tag;
    if (Test.directional && Test.genStop && ! Test.brokenDep) {
      tempDependency.leftHeaded = true;
      score += rawScore(tempDependency);
    }
    if (Test.genStop) {
      tempDependency.leftHeaded = false;;
      score += rawScore(tempDependency);
    }
    return score;
    }*/

  public double score(Dependency dependency) {
    short hTBackup = dependency.head.tag;
    short aTBackup = dependency.arg.tag;

    dependency.head.tag = (short) tagBin(dependency.head.tag);
    dependency.arg.tag = (short) tagBin(dependency.arg.tag);

    double s = scoreTB(dependency);

    dependency.head.tag = hTBackup;
    dependency.arg.tag = aTBackup;

    return s;
  }

  public double scoreTB(Dependency dependency) {
    boolean verbose = false;
    if (!Options.get().directional)
      dependency.leftHeaded = false;
    if (verbose)
      System.out.println("Generating " + dependency);
    short distance = dependency.distance;
    boolean leftHeaded = dependency.leftHeaded;
    int hW = dependency.head.word;
    int aW = dependency.arg.word;
    IntTaggedWord aTW = dependency.arg;
    IntTaggedWord hTW = dependency.head;

    dependency.distance = valenceBin(distance);

    double c_aTW_hTWd = argCounter.getCount(dependency);
    dependency.arg.word = -1;
    double c_aT_hTWd = argCounter.getCount(dependency);
    dependency.arg.word = aW;
    dependency.arg = wildTW;
    double c_hTWd = argCounter.getCount(dependency);
    dependency.arg = aTW;

    dependency.head.word = -1;
    double c_aTW_hTd = argCounter.getCount(dependency);
    dependency.arg.word = -1;
    double c_aT_hTd = argCounter.getCount(dependency);
    dependency.arg.word = aW;
    dependency.arg = wildTW;
    double c_hTd = argCounter.getCount(dependency);
    dependency.arg = aTW;
    dependency.head.word = hW;

    dependency.distance = distanceBin(distance);
    dependency.arg = stopTW;
    double c_stop_hTWds = stopCounter.getCount(dependency);
    dependency.head.word = -1;
    double c_stop_hTds = stopCounter.getCount(dependency);
    dependency.head.word = hW;
    dependency.arg = wildTW;
    double c_hTWds = stopCounter.getCount(dependency);
    dependency.head.word = -1;
    double c_hTds = stopCounter.getCount(dependency);
    dependency.head.word = hW;
    dependency.arg = aTW;

    dependency.head = wildTW;
    dependency.leftHeaded = false;
    dependency.distance = -1;
    double c_aTW = argCounter.getCount(dependency);
    dependency.arg.word = -1;
    double c_aT = argCounter.getCount(dependency);
    dependency.arg.word = aW;
    dependency.leftHeaded = leftHeaded;
    dependency.head = hTW;
    dependency.distance = distance;

    // decide whether to generate something

    double p_stop_hTWds = (c_hTWds > 0.0 ? c_stop_hTWds / c_hTWds : 0.0);
    double p_stop_hTds = (c_hTds > 0.0 ? c_stop_hTds / c_hTds : 1.0);

    double pb_stop_hTWds = (c_stop_hTWds + smooth[2] * p_stop_hTds) / (c_hTWds + smooth[2]);

    if (verbose) {
      System.out.println("  c_stop_hTWds: " + c_stop_hTWds + "; c_hTWds: " + c_hTWds +
              "; c_stop_hTds: " + c_stop_hTds + "; c_hTds: " + c_hTds);
      System.out.println("  Generate STOP prob: " + pb_stop_hTWds);
    }

    if (aW == -2) {
      // did we generate stop?
      if (rootTW(hTW))
        return 0.0;
      return Test.depWeight * Math.log(pb_stop_hTWds);
    }

    double pb_go_hTWds = 1.0 - pb_stop_hTWds;

    if (rootTW(hTW)) {
      pb_go_hTWds = 1.0;
    }

    // generate the argument

    // do the magic
    double p_aTW_hTWd = (c_hTWd > 0.0 ? c_aTW_hTWd / c_hTWd : 0.0);
    double p_aTW_hTd = (c_hTd > 0.0 ? c_aTW_hTd / c_hTd : 0.0);
    double p_aT_hTWd = (c_hTWd > 0.0 ? c_aT_hTWd / c_hTWd : 0.0);
    double p_aT_hTd = (c_hTd > 0.0 ? c_aT_hTd / c_hTd : 0.0);
    double p_aTW_aT = (c_aTW > 0.0 ? c_aTW / c_aT : 1.0);
    //double p_hd_hTW = (c_hTW>0.0 ? c_hTWd/c_hTW : 0.0);
    //double p_hd_hT = (c_hT>0.0 ? c_hTd/c_hT : 0.0);

    double pb_aTW_hTWd = (c_aTW_hTWd + smooth[0] * p_aTW_hTd) / (c_hTWd + smooth[0]);
    double pb_aT_hTWd = (c_aT_hTWd + smooth[1] * p_aT_hTd) / (c_hTWd + smooth[1]);

    //double nScore = Math.log(p_aT_hTd);

    double score = (Math.log(interp[0] * pb_aTW_hTWd +
            (1.0 - interp[0]) * p_aTW_aT * pb_aT_hTWd) +
            Math.log(pb_go_hTWds));

    if (verbose) {
      NumberFormat nf = NumberFormat.getNumberInstance();
      nf.setMaximumFractionDigits(2);
      System.out.println("  c_aTW_hTWd: " + c_aTW_hTWd + "; c_aT_hTWd: " + c_aT_hTWd + "; c_hTWd: " + c_hTWd);
      System.out.println("  c_aTW_hTd: " + c_aTW_hTd + "; c_aT_hTd: " + c_aT_hTd + "; c_hTd: " + c_hTd);
      System.out.println("  Generated with pb_go_hTWds: " + nf.format(Math.log(pb_go_hTWds)) + " pb_aTW_hTWd: " + nf.format(Math.log(pb_aTW_hTWd)) + " p_aTW_aT: " + nf.format(Math.log(p_aTW_aT)) + " pb_aT_hTWd: " + nf.format(Math.log(pb_aT_hTWd)));
      System.out.println("  NoDist log score: " + score);
    }

    if (Test.prunePunc && pruneTW(aTW))
      return 0.0;

    if (Double.isNaN(score))
      score = Double.NEGATIVE_INFINITY;

    //if (Test.rightBonus && ! dependency.leftHeaded)
    //  score -= 0.2;

    if (score < -100)
      score = Double.NEGATIVE_INFINITY;

    return Test.depWeight * score;
  }

  private Dependency d = new Dependency(-2, -2, -2, -2, false, 0);

  public double score(int headWord, int headTag, int argWord, int argTag, boolean leftHeaded, int dist) {
    d.head.word = headWord;
    d.head.tag = (short) headTag;
    d.arg.word = argWord;
    d.arg.tag = (short) argTag;
    d.leftHeaded = leftHeaded;
    d.distance = (short) dist;
    return score(d);
  }

  public double scoreTB(int headWord, int headTag, int argWord, int argTag, boolean leftHeaded, int dist) {
    d.head.word = headWord;
    d.head.tag = (short) headTag;
    d.arg.word = argWord;
    d.arg.tag = (short) argTag;
    d.leftHeaded = leftHeaded;
    d.distance = (short) dist;
    return scoreTB(d);
  }

  void initTagBins() {
    Numberer tagBinNumberer = new Numberer();
    tagBin = new int[numTags];
    Arrays.fill(tagBin, -1);
    for (int t = 0; t < numTags; t++) {
      String tagStr = (String) tagNumberer.object(t);
      String binStr = tagStr;
      if (tagProjection != null)
        binStr = tagProjection.project(tagStr);
      int bin = tagBinNumberer.number(binStr);
      //      System.err.println("Mapped "+tagStr+" to "+binStr);
      tagBin[t] = bin;
    }
    numTagBins = tagBinNumberer.total();
    //    System.err.println("tags: "+numTags+" bins "+numTagBins);
  }

  public DependencyGrammar(int numWords, int numTags) {
    this(new TestTagProjection());
  }

  public DependencyGrammar(){}
  public DependencyGrammar(TagProjection tagProjection) {
    //this.numWords = numWords;
    //this.numTags = numTags;
    //headArgCounter = new Counter();
    //headDistCounter = new Counter();
    //headCounter = new Counter();
    //lheadCounter = new Counter();
    //rheadCounter = new Counter();
    this.tagProjection = tagProjection;
    argCounter = new Counter();
    stopCounter = new Counter();
    tagNumberer = Numberer.getGlobalNumberer("tags");
    wordNumberer = Numberer.getGlobalNumberer("words");
    numTags = tagNumberer.total();
    numWords = wordNumberer.total();
    initTagBins();
  }

  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject();
    argCounter = new Counter();
    stopCounter = new Counter();
    for (Iterator dI = coreDependencies.keySet().iterator(); dI.hasNext();) {
      Dependency d = (Dependency) dI.next();
      double count = coreDependencies.getCount(d);
      expandDependency(d, count);
    }
    expandDependencyMap = null;
  }

  public Counter getCoreDependencies() {
    return coreDependencies;
  }

  /**
   * Populates data in this DependencyGrammar from the character stream
   * given by the Reader r.
   */
  public void readData(BufferedReader in) throws IOException {
    final String LEFT = "left";
    String line, headString, argString, dirString, distanceString, countString;
    int lineNum = 1;
    StringTokenizer tok;
    Dependency d;
    boolean leftHeaded;
// all lines have one rule per line
    line = in.readLine();
    while (line != null && !line.startsWith("**********")) {
      try {
        tok = new StringTokenizer(line, "\t"); // break on tabs only
        headString = tok.nextToken();
        tok.nextToken();
        argString = tok.nextToken();
        dirString = tok.nextToken();
        if (dirString.equals(LEFT))
          leftHeaded = true;
        else
          leftHeaded = false;
        distanceString = tok.nextToken();
        countString = tok.nextToken();
        d = new Dependency(new IntTaggedWord(headString), new IntTaggedWord(argString),
                leftHeaded, Integer.parseInt(distanceString));
        addRule(d, Double.parseDouble(countString));
      } catch (Exception e) {
        e.printStackTrace();
        throw new IOException("Error on line " + lineNum);
      }
//      System.out.println("read line " + lineNum + ": " + line);
      lineNum++;
      line = in.readLine();
    }
  }

  /**
   * Writes out data from this Object to the Writer w.
   */
  public void writeData(Writer w) throws IOException {
    PrintWriter out = new PrintWriter(w);
// all lines have one rule per line
    for (Iterator dI = coreDependencies.keySet().iterator(); dI.hasNext();) {
      Dependency d = (Dependency) dI.next();
      double count = coreDependencies.getCount(d);
      out.println(d + "\t" + count);
    }
    out.flush();
  }

  private static final long serialVersionUID = 1L;

} // end class DependencyGrammar

