// StanfordLexicalizedParser -- a probabilistic lexicalized NL CFG parser
// Copyright (c) 2002, 2003 Leland Stanford Junior University
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//
// For more information, bug reports, fixes, contact:
//    Christopher Manning
//    Dept of Computer Science, Gates 4A
//    Stanford CA 94305-9040
//    USA
//    parser-support@lists.stanford.edu
//    http://nlp.stanford.edu/downloads/lex-parser.shtml

package edu.stanford.nlp.parser.lexparser;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.zip.*;
import java.util.regex.*;

import edu.stanford.nlp.parser.*;
import edu.stanford.nlp.trees.*;
import edu.stanford.nlp.util.*;
import edu.stanford.nlp.io.NumberRangeFileFilter;
import edu.stanford.nlp.process.*;
import edu.stanford.nlp.dbm.*;
import edu.stanford.nlp.objectbank.TokenizerFactory;
import edu.stanford.nlp.fsm.ExactGrammarCompactor;

/** A reasonably good lexicalized PCFG parser.  It does a product-of-experts
 *  model of plain PCFG parsing and lexicalized dependency parsing.  Or it
 *  can do unlexicalized PCFG parsing by using just that component parser.
 *  Note that training requires a lot of memory to run.  Try -mx1500m.
 *  See the package documentation for more details and examples of use.
 *  See the main method documentation for details of invoking the parser.
 *
 *  @author Dan Klein (original version)
 *  @author Christopher Manning (better features, ParserParams, serialization)
 *  @author Roger Levy (internationalization)
 *  @author Teg Grenager (grammar compaction, etc., tokenization, etc.)
 */
public class LexicalizedParser implements ViterbiParser, Function {

  private boolean fallbackToPCFG = false;

  protected ExhaustivePCFGParser pparser;
  protected ExhaustiveDependencyParser dparser;
  protected BiLexPCFGParser bparser;
  protected TreeTransformer debinarizer;
  private SubcategoryStripper subcategoryStripper;

  private transient ParserData pd;
  private TreebankLangParserParams tlpParams =
          new EnglishTreebankParserParams();

  private static final String SERIALIZED_PARSER_PROPERTY =
          "edu.stanford.nlp.SerializedLexicalizedParser";
  private static final String DEFAULT_PARSER_LOC =
          "/u/nlp/data/lexparser/serializedParser.gz";


  /** Allows the caller to specify a TreebankLangParserParams to use.
   *  @param tlpp The one to use
   */
  public void setTreebankLangParserParams(TreebankLangParserParams tlpp) {
    tlpParams = tlpp;
  }


  /** Converts a Sentence/List into a Tree.
   *  If it can't be parsed, it is made into a trivial tree in which each
   *  word is attached to a start nonterminal.
   *  @param in The input Sentence/List
   *  @return A Tree that is the parse tree for the sentence.  If the parser
   *          fails, a new Tree is synthesized which attaches all words to the
   *          root.
   *  @throws IllegalArgumentException If argument isn't a List
   */
  public Object apply(Object in) {
    if ( ! (in instanceof List)) {
      throw new IllegalArgumentException("Can only parse Sentence/List");
    }
    List lst = (List) in;
    try {
      if (parse(lst)) {
        return getBestParse();
      }
    } catch (Exception e) {
      e.printStackTrace();
    }
    // if can't parse or exception, fall through
    TreeFactory lstf = new LabeledScoredTreeFactory();
    String root = Options.get().langpack().startSymbol();
    List lst2 = new Sentence();
    for (Iterator it = lst.iterator(); it.hasNext(); ) {
      String s = it.next().toString();
      Tree t = lstf.newLeaf(s);
      lst2.add(t);
    }
    return lstf.newTreeNode(root, lst2);
  }


  /** Parse a Sentence.
   *  @return true iff it could be parsed
   */
  public boolean parse(Sentence sentence) {
    return parse((List) sentence);
  }


  /** Parse a Sentence.
   *  This hasn't yet been implemented.  At present the goal is ignored.
   *  @return true iff it could be parsed
   */
  public boolean parse(Sentence sentence, String goal) {
    return parse((List) sentence);
  }


  private static boolean parseSucceeded = false;

  /** Parse a sentence represented as a List.
   *  @param sentence The sentence to parse
   *  @return true Iff the sentence was accepted by the grammar
   *  @throws UnsupportedOperationException If the Sentence is too long or
   *        otherwise fails for resource reasons
   */
  public boolean parse(List sentence) {
    TreePrint treePrint = Test.treePrint();
    PrintWriter pwOut = tlpParams.pw();
    PrintWriter pwErr = tlpParams.pw(System.err);
    parseSucceeded = false;
    Sentence sentenceB = new Sentence(sentence);
    sentenceB.add(Lexicon.BOUNDARY);
    int length = sentenceB.length();
    if (length > Test.maxLength + 1) {  // + 1 for boundary symbol
      throw new UnsupportedOperationException("Sentence too long: " + length);
    }
    if (Options.get().doPCFG) {
      if (!pparser.parse(sentenceB))
        return parseSucceeded;
      if (Test.verbose && Options.get().doDep) {
        System.out.println("PParser output");
        treePrint.printTree(debinarizer.transformTree(pparser.getBestParse(), null), pwOut); //combo parse goes to System.out
        //(subcategoryStripper.transformTree(debinarizer.transformTree(pparser.getBestParse(),null),null)).pennPrint();
      }
    }
    if (Options.get().doDep) {
      dparser.parse(sentenceB);
      if (Test.verbose) {
        System.out.println("DParser output");
        treePrint.printTree(dparser.getBestParse(), pwOut); //combo parse goes to System.out
      }
    }
    if (Options.get().doPCFG &&
            Options.get().doDep) {
      if (!bparser.parse(sentenceB))
        return parseSucceeded;
      else
        parseSucceeded = true;
    }
    return true;//parseSucceeded;
  }


  /** Return the best parse of the sentence most recently parsed.
   *  @return The best tree
   *  @throws NoSuchElementException If no previously successfully parsed
   *            sentence
   */
  public Tree getBestParse() {
    if (!parseSucceeded || bparser == null) {
      if (!pparser.hasParse() ||
              !fallbackToPCFG)
        throw new NoSuchElementException();
      else
        return getBestPCFGParse();
    }
    Tree binaryTree = bparser.getBestParse();
    return subcategoryStripper.transformTree(debinarizer.transformTree(
            binaryTree, null), null);
  }

  public Tree getBestPCFGParse() {
    return getBestPCFGParse(true);
  }

  public Tree getBestPCFGParse(boolean stripSubcategories) {
    Tree binaryTree = pparser.getBestParse();
    if (binaryTree == null) return null;
    Tree t = debinarizer.transformTree(binaryTree, null);
    if (stripSubcategories)
      t = subcategoryStripper.transformTree(t, null);
    return t;
  }

  public double getPCFGScore(String goalStr) {
    if (!pparser.stateNumberer.hasSeen(goalStr))
      return Double.NEGATIVE_INFINITY;
    pparser.setGoalString(goalStr);
    return pparser.getLogScore();
  }

  public Tree getBestDependencyParse() {
    return (dparser != null ? dparser.getBestParse() : null);
  }


  /**
   * Construct a new LexicalizedParser object from a previously serialized
   * grammar read from a property
   * <code>edu.stanford.nlp.SerializedLexicalizedParser</code>,
   * or a default file location.
   */
  public LexicalizedParser() {
    String source =
            System.getProperty(SERIALIZED_PARSER_PROPERTY);
    if (source == null) {
      source = DEFAULT_PARSER_LOC;
    }
    pd = getParserDataFromSerializedFile(source);
    makeParsers(pd);
  }


  /**
   * Construct a new LexicalizedParser.  This loads a grammar that
   * was previously assembled and stored.
   * @throws IllegalArgumentException If parser data cannot be loaded
   */
  public LexicalizedParser(String parserFileOrUrl) {
    System.err.print("Loading parser from file " + parserFileOrUrl);
    pd = getParserDataFromSerializedFile(parserFileOrUrl);
    if (pd == null) {
      pd = getParserDataFromTextFile(parserFileOrUrl);
    }
    makeParsers(pd);
    System.err.println("done.");
  }

  /**
   * Construct a new LexicalizedParser.  This loads a grammar that
   * was previously assembled and stored.
   * @throws IllegalArgumentException If parser data cannot be loaded
   */
  public LexicalizedParser(String parserFileOrUrl, boolean isTextGrammar) {
    System.err.print("Loading parser from file " + parserFileOrUrl);
    if (isTextGrammar) {
      pd = getParserDataFromTextFile(parserFileOrUrl);
    } else {
      pd = getParserDataFromSerializedFile(parserFileOrUrl);
    }
    makeParsers(pd);
    System.err.println("done.");
  }


  /**
   * Construct a new LexicalizedParser.  This loads a grammar that
   * was previously assembled and stored.
   * @param maxLeng Maximum sentence length that you want the parser to be
   *    able to parse (this effects memory consumption)
   * @throws IllegalArgumentException If parser data cannot be loaded
   */
  public LexicalizedParser(String serializedFileOrUrl, int maxLeng) {
    Test.maxLength = maxLeng;
    pd = getParserDataFromSerializedFile(serializedFileOrUrl);
    makeParsers(pd);
  }


  /**
   * Construct a new LexicalizedParser object from a previously assembled
   * grammar.
   * @param pd A <code>ParserData</code> object (not <code>null</code>)
   */
  public LexicalizedParser(ParserData pd) {
    this.pd = pd;
    makeParsers(pd);
  }


  /**
   * Construct a new LexicalizedParser object from a previously assembled
   * grammar read from an InputStream.
   * One (ParserData) object is read from the stream.  It isn't closed.
   * @param in The ObjectInputStream
   */
  public LexicalizedParser(ObjectInputStream in) throws Exception {
    this((ParserData) in.readObject());
  }


  /**
   * Construct a new LexicalizedParser object from a previously assembled
   * grammar read from an InputStream.
   * One (ParserData) object is read from the stream.  It isn't closed.
   * @param in The ObjectInputStream
   * @param maxLeng Maximum sentence length that you want the parser to be
   *    able to parse (this effects memory consumption)
   */
  public LexicalizedParser(ObjectInputStream in, int maxLeng)
          throws Exception {
    this(in);
    Test.maxLength = maxLeng;
  }


  /**
   * Construct a new LexicalizedParser by training from treebank files.
   * @param treebankPath a <code>String</code> value
   * @param filt a <code>FileFilter</code> value. This may be
   *       <code>null</code> if no filtering of selected files is needed.
   */
  public LexicalizedParser(String treebankPath, FileFilter filt,
                           TreebankLangParserParams tlpParams,
                           GrammarCompactor compactor) {
    this.tlpParams = tlpParams;
    pd = getParserDataFromTreebank(treebankPath, filt, compactor);
    makeParsers(pd);
  }

  public LexicalizedParser(String treebankPath, FileFilter filt,
                           TreebankLangParserParams tlpParams) {
    this(treebankPath, filt, tlpParams, null);
  }

  /**
   * Construct a new LexicalizedParser by training from treebank files.
   * @param treebankPath a <code>String</code> value
   */
  public LexicalizedParser(String treebankPath,
                           TreebankLangParserParams tlpParams,
                           GrammarCompactor compactor) {
    this(treebankPath, null, tlpParams, compactor);
  }


  /**
   * Construct a new LexicalizedParser.
   * @param treebankPath a <code>String</code> value
   * @param filt a <code>FileFilter</code> value
   * @param maxLeng The maximum length sentences to be able to parser.
   *           A large value for this requires a great deal of memory (and
   *           time) for parsing, but allows parsing longer sentences.
   * @param tlpParams The Treebank parameters class for different languages
   */
  public LexicalizedParser(String treebankPath, FileFilter filt, int maxLeng,
                           TreebankLangParserParams tlpParams,
                           GrammarCompactor compactor) {
    this.tlpParams = tlpParams;
    Test.maxLength = maxLeng;
    pd = getParserDataFromTreebank(treebankPath, filt, compactor);
    makeParsers(pd);
  }


  /**
   * Construct a new LexicalizedParser.
   * @param treebankPath a <code>String</code> value
   * @param filt a <code>FileFilter</code> value
   * @param maxLeng The maximum length sentences to be able to parser.
   *           A large value for this requires a great deal of memory (and
   *           time) for parsing, but allows parsing longer sentences.
   */
  public LexicalizedParser(String treebankPath, FileFilter filt, int maxLeng,
                           GrammarCompactor compactor) {
    Test.maxLength = maxLeng;
    pd = getParserDataFromTreebank(treebankPath, filt, compactor);
    makeParsers(pd);
  }

  public LexicalizedParser(String treebankPath, FileFilter filt, int maxLeng) {
    this(treebankPath, filt, maxLeng, null);
  }


  public ParserData parserData() {
    return pd;
  }

  static void saveParserDataToSerialized(ParserData pd, String filename) {
    try {
      System.err.print("Writing parser in serialized format to file " + filename + " ");
      ObjectOutputStream out;
      if (filename.endsWith(".gz")) {
        // it's faster to do the buffering _outside_ the gzipping as here
        out = new ObjectOutputStream(new BufferedOutputStream(
                new GZIPOutputStream(new FileOutputStream(filename))));
      } else {
        out = new ObjectOutputStream(
                new BufferedOutputStream(new FileOutputStream(filename)));
      }
      out.writeObject(pd);
      out.close();
      System.err.println("done.");
    } catch (IOException ioe) {
      ioe.printStackTrace();
    }
  }

  static void saveParserDataToText(ParserData pd, String filename) {
    try {
      System.err.print("Writing parser in text grammar format to file " + filename);
      OutputStream os = null;
      if (filename.endsWith(".gz")) {
        // it's faster to do the buffering _outside_ the gzipping as here
        os = new BufferedOutputStream(new GZIPOutputStream(new FileOutputStream(filename)));
      } else {
        os = new BufferedOutputStream(new FileOutputStream(filename));
      }
      PrintWriter out = new PrintWriter(os);
      String prefix = "**********BEGIN ";
      String suffix = "**********";
      out.println(prefix + "OPTIONS" + suffix);
      if (pd.pt != null) {
        pd.pt.writeData(out);
      }
      System.err.print(".");
      out.println(prefix + "LEXICON" + suffix);
      if (pd.lex != null) {
        pd.lex.writeData(out);
      }
      System.err.print(".");
      out.println(prefix + "UNARY GRAMMAR" + suffix);
      if (pd.ug != null) {
        pd.ug.writeData(out);
      }
      System.err.print(".");
      out.println(prefix + "BINARY GRAMMAR" + suffix);
      if (pd.bg != null) {
        pd.bg.writeData(out);
      }
      System.err.print(".");
      out.println(prefix + "DEPENDENCY GRAMMAR" + suffix);
      if (pd.dg != null) {
        pd.dg.writeData(out);
      }
      System.err.print(".");
      out.flush();
      out.close();
      System.err.println("done.");
    } catch (IOException e) {
      System.err.println("Trouble saving parser data to ASCII format.");
      e.printStackTrace();
    }
  }


  protected static ParserData getParserDataFromTextFile(String textFileOrUrl) {
    try {
      InputStream is;
      BufferedReader in;
      if (textFileOrUrl.startsWith("http://")) {
        URL u = new URL(textFileOrUrl);
        URLConnection uc = u.openConnection();
        is = uc.getInputStream();
      } else {
        is = new FileInputStream(textFileOrUrl);
      }
      if (textFileOrUrl.endsWith(".gz")) {
        in = new BufferedReader(new InputStreamReader(new GZIPInputStream(is)));
      } else {
        in = new BufferedReader(new InputStreamReader(is));
      }
      System.err.print(" in text grammar format");
      Timing.startTime();
      in.readLine();
      Options opt = Options.get();
      opt.readData(in);
      System.err.print(".");
      Lexicon lex = new Lexicon();
      lex.readData(in);
      System.err.print(".");
      UnaryGrammar ug = new UnaryGrammar(opt.numStates);
      ug.readData(in);
      System.err.print(".");
      BinaryGrammar bg = new BinaryGrammar(opt.numStates);
      bg.readData(in);
      System.err.print(".");
      DependencyGrammar dg = new DependencyGrammar(opt.numWords, opt.numTags);
      dg.readData(in);
      System.err.print(".");
      Map numbs = Numberer.getNumberers();
      in.close();
      return new ParserData(lex, bg, ug, dg, numbs, Options.get());
    } catch (IOException e) {
      e.printStackTrace();
    }
    return null;
  }

  protected static ParserData getParserDataFromSerializedFile(String serializedFileOrUrl) {
    ParserData pd = null;
    try {
      ObjectInputStream in;
      InputStream is;
      if (serializedFileOrUrl.startsWith("http://")) {
        URL u = new URL(serializedFileOrUrl);
        URLConnection uc = u.openConnection();
        is = uc.getInputStream();
      } else {
        is = new FileInputStream(serializedFileOrUrl);
      }
      if (serializedFileOrUrl.endsWith(".gz")) {
        // it's faster to do the buffering _outside_ the gzipping as here
        in = new ObjectInputStream(new BufferedInputStream(new GZIPInputStream(
                is)));
      } else {
        in = new ObjectInputStream(new BufferedInputStream(is));
      }
      Timing.startTime();
      pd = (ParserData) in.readObject();
      System.err.print(" in serialized format...");
      in.close();
      return pd;
    } catch (StreamCorruptedException sce) {
      // suppress error message, on the assumption that we've really got
      // a text grammar, and that'll be tried next
    } catch (Exception e) {
      e.printStackTrace();
    }
    return null;
  }


  private static void printOptions(boolean train) {
    Options.get().display();
    if (train) {
      Train.display();
    } else {
      Test.display();
    }
    Options.get().tlpParams.display();
  }


  protected final ParserData getParserDataFromTreebank(String treebankPath,
                                                       FileFilter filt, GrammarCompactor compactor) {
    System.err.println("Training a parser from treebank dir: " + treebankPath);
    System.out.println("Currently " + new Date());
    printOptions(true);
    Timing.startTime();
    // setup tree transforms
    Treebank trainTreebank = tlpParams.memoryTreebank();
    System.out.print("Reading trees...");
    if (filt == null) {
      trainTreebank.loadPath(treebankPath);
    } else {
      trainTreebank.loadPath(treebankPath, filt);
    }
    Timing.tick("done.");
    System.out.print("Binarizing trees...");
    TreeTransformer binarizer = null;
    if (!Train.leftToRight) {
      binarizer = BinarizerFactory.getHeadRespectingBinarizer(tlpParams.headFinder(), !Train.outsideFactor());
    } else {
      binarizer = BinarizerFactory.getHeadRespectingBinarizer(new LeftHeadFinder(), !Train.outsideFactor());
    }
    List binaryTrainTrees = new ArrayList();
    if (Train.hSelSplit) {
      ((BinarizerFactory.TreeAnnotatorAndBinarizer) binarizer).setCountStates(true);
      for (Iterator tI = trainTreebank.iterator(); tI.hasNext();) {
        Tree tree = (Tree) tI.next();
        if (Train.collinsPunc)
          tree = CollinsPuncTransformer.transformTree(tree);
        tree = binarizer.transformTree(tree, null);
      }
      ((BinarizerFactory.TreeAnnotatorAndBinarizer) binarizer).setCountStates(false);
    }
    for (Iterator tI = trainTreebank.iterator(); tI.hasNext();) {
      Tree tree = (Tree) tI.next();
      if (Train.collinsPunc)
        tree = CollinsPuncTransformer.transformTree(tree);
      tree = binarizer.transformTree(tree, null);

      binaryTrainTrees.add(tree);
    }
    Timing.tick("done.");
    // extract grammars
    //Extractor bgExtractor = new SmoothedBinaryGrammarExtractor();
    Extractor bgExtractor = new BinaryGrammarExtractor();
    // Extractor lexExtractor = new LexiconExtractor();
    Extractor dgExtractor = new DependencyGrammarExtractor();
    //TreeExtractor uwmExtractor = new UnknownWordModelExtractor(binaryTrainTrees.size());
    Extractor cdExtractor = new CorrectDependencyExtractor();
    BinaryGrammar bg = null;
    UnaryGrammar ug = null;
    System.out.print("Extracting PCFG...");
    Pair bgug = (Pair) bgExtractor.extract(binaryTrainTrees);
    Timing.tick("done.");
    if (compactor != null) {
      System.out.print("Compacting grammar...");
      bgug = compactor.compactGrammar(bgug);
      Timing.tick("done.");
    }
    System.out.print("Compiling grammar...");
    bg = (BinaryGrammar) bgug.second;
    bg.splitRules();
    ug = (UnaryGrammar) bgug.first;
    ug.purgeRules();
    Timing.tick("done");
    System.out.print("Extracting Lexicon...");
    Lexicon lex = Options.get().tlpParams.lex();
    lex.train(binaryTrainTrees);
    Timing.tick("done.");
    System.out.print("Tuning Lexicon...");
    lex.tune(binaryTrainTrees);
    Timing.tick("done.");

    DependencyGrammar dg = null;
    if (Options.get().doDep) {
      System.out.print("Extracting Dependencies...");
      dg = (DependencyGrammar) dgExtractor.extract(binaryTrainTrees);
      Timing.tick("done.");
      //System.out.println(dg);
      //System.out.print("Extracting Unknown Word Model...");
      //UnknownWordModel uwm = (UnknownWordModel)uwmExtractor.extract(binaryTrainTrees);
      //Timing.tick("done.");
      System.out.print("Extracting Tune Dependencies...");
      List tuneDeps = (List) cdExtractor.extract(binaryTrainTrees);
      Timing.tick("done.");
      System.out.print("Tuning Dependency Model...");
      dg.tune(tuneDeps);
      Timing.tick("done.");
    }
    Map numbs = Numberer.getNumberers();

    int numStates = Numberer.getGlobalNumberer("states").total(); // TEG debug
    System.out.println("numStates: " + numStates); // TEG debug

    System.err.println("Done training parser.");
    return new ParserData(lex, bg, ug, dg, numbs, Options.get());
  }


  protected final void makeParsers(ParserData pd) {
    if (pd == null) {
      throw new IllegalArgumentException("Error loading parser data: pd null");
    }
    BinaryGrammar bg = pd.bg;
    UnaryGrammar ug = pd.ug;
    Lexicon lex = pd.lex;
    DependencyGrammar dg = pd.dg;
    Numberer.setNumberers(pd.numbs);
    Options.set(pd.pt);
    pparser = new ExhaustivePCFGParser(bg, ug, lex);
    if (Options.get().doDep) {
      dparser = new ExhaustiveDependencyParser(dg, lex);

      Scorer scorer = new TwinScorer(pparser, dparser);
      //Scorer scorer = parser;
      bparser = new BiLexPCFGParser(scorer, pparser, dparser, bg, ug, dg, lex);
      bparser.setMaxItems(200000);
    }
    fallbackToPCFG = true;

    debinarizer = BinarizerFactory.getDebinarizer();
    subcategoryStripper = new SubcategoryStripper();
  }


  /** returns the input sentence for the parser */
  static Sentence getInputSentence(Tree t) {
    if (Test.forceTags) {
      if (Test.preTag) // doesn't yet support
        throw new RuntimeException("Sorry -- haven't implemented support for external tagger yet.");
      else
        return t.taggedYield();
    } else
      return t.yield();
  }

  // helper function
  private static int numSubArgs(String[] args, int index) {
    int i = index;
    while (i + 1 < args.length && args[i + 1].charAt(0) != '-') {
      i++;
    }
    return i - index;
  }

  public void testGrammarCoverage(Treebank testTreebank) {
    TreePrint treePrint = Test.treePrint();
    PrintWriter pwOut = tlpParams.pw();
    PrintWriter pwErr = tlpParams.pw(System.err);
    TreeTransformer tc = tlpParams.collinizer();
    TreeTransformer annotator = new BinarizerFactory.TreeAnnotator(tlpParams.headFinder());
    CategoryWordTag.printWordTag = false;
    for (Iterator it = testTreebank.iterator(); it.hasNext();) {
      Tree goldTree = (Tree) it.next();
      // now annotate it
      Tree annotatedTree = annotator.transformTree(goldTree, null);
      // get all local trees and iterate through them
      List localTrees = annotatedTree.subTreeList(); // list of pointers to each node in annotatedTree
      for (Iterator iter = localTrees.iterator(); iter.hasNext();) {
        Tree tree = (Tree) iter.next();
        // find the score of the local tree
        if (tree.isPhrasal()) {
          double score = scoreLocalTreeWithGrammar(tree);
          ((LabeledScoredTreeNode) tree).setScore(score);
        }
      }
      annotatedTree.pennPrint(); // this should print scores, right?
      System.out.println();
    }
    CategoryWordTag.printWordTag = true;
  }

  private double scoreLocalTreeWithGrammar(Tree localTree) {
    return 8.2171; // for now to test
  }


  /** Evaluates the performance of the parser on a test treebank.
   *  Note that this routine prints material to stdout and stderr for
   *  each tree that it parses.
   *  @param testTreebank The Treebank to test the parser on.
   *  @return The labeled constituent F1 score (as a percentage).
   */
  public double testOnTreebank(Treebank testTreebank) {
    System.err.println("Testing on treebank");
    Timing.startTime();
    TreePrint treePrint = Test.treePrint();
    PrintWriter pwOut = tlpParams.pw();
    PrintWriter pwErr = tlpParams.pw(System.err);
    TreeTransformer tc = tlpParams.collinizer();
    TreeTransformer br = new BoundaryRemover();
    TreeFactory lstf = new LabeledScoredTreeFactory();
    LabeledConstituentEval pcfgPE = new LabeledConstituentEval("pcfg LP/LR");
    DependencyEval depDE = new DependencyEval("depend DE");
    LabeledConstituentEval comboPE = new LabeledConstituentEval("factor LP/LR");
    // RuleErrorEval ruleErrorEval = new RuleErrorEval("Rule under/over");
    TaggingEval comboTE = new TaggingEval("Tag", pd.lex);
    TreeTransformer binarizerOnly = BinarizerFactory.getHeadRespectingBinarizer(tlpParams.headFinder(), false, false);
    for (Iterator it = testTreebank.iterator(); it.hasNext();) {
      Tree goldTree = (Tree) it.next();
//System.out.println(goldTree.label().getClass());//debugging
//System.out.println(Test.forceTags);
      Sentence s = getInputSentence(goldTree);
//System.out.println("is it a taggedword?" + (s.get(0) instanceof TaggedWord));
      long time = System.currentTimeMillis();
// REMOVE
//if (goldTree.yield().length() != Test.maxLength)
//  continue;
      pwErr.println("Parsing [len. " + s.length() + "]: " + s);
      Tree tree = null;
      try {
        if (!parse(s)) {
          pwErr.println("Sentence couldn't be parsed by grammar.");
        } else {
          tree = getBestParse();
          if (Test.verbose) {
            pwOut.println("ComboParser best");
          }
          treePrint.printTree(tree, pwOut); //combo parse goes to System.out
          if (Test.printAllBestParses) {
            List parses = pparser.getBestParses();
            int sz = parses.size();
            if (sz > 1) {
              pwOut.println("There were " + sz + " best parses.");
              Tree transGoldTree = tc.transformTree(goldTree, null);
              for (Iterator itb = parses.iterator(); itb.hasNext();) {
                Tree tb = (Tree) itb.next();
                Tree tbd = debinarizer.transformTree(tb, null);
                tbd = subcategoryStripper.transformTree(tbd, null);
                Tree tbtr = tc.transformTree(tbd, null);
                pwOut.println("Tree size = " + tbtr.size() + "; depth = " +
                              tbtr.depth());
                tbtr.pennPrint(pwOut);
                pcfgPE.evaluate(tbtr, transGoldTree, pwOut);
              }
            }
          }
          // if (pparser != null)
          //   pwErr.println("Correct parse\tScore: " + pparser.scoreBinarizedTree(binarizer.transformTree(goldTree,null),true));
          // else
          if (Test.verbose) {
            pwOut.println("Correct parse");
            treePrint.printTree(goldTree, pwOut);
          }
        }
      } catch (UnsupportedOperationException uoe) {
        pwErr.println("Sentence too long");
      }
      long elapsed = System.currentTimeMillis() - time;
      pwErr.println("Time: " + (elapsed / 1000) + " sec.");
      if (tree != null) {
        Tree transGoldTree = tc.transformTree(goldTree, null);
        Tree treePCFG = getBestPCFGParse();
        pcfgPE.evaluate(tc.transformTree(treePCFG, null), transGoldTree, pwErr);
        // ruleErrorEval.evaluate(tc.transformTree(treePCFG, null), transGoldTree, pwErr);
        Tree treeDep = getBestDependencyParse();
        if (treeDep != null)
          depDE.evaluate(treeDep, binarizerOnly.transformTree(goldTree, null), pwErr);
        comboPE.evaluate(tc.transformTree(tree, null), transGoldTree, pwErr);
        comboTE.evaluate(tree, br.transformTree(goldTree, lstf), pwErr);
      }
      pwErr.println();
    } // for tree iterator
    Timing.tick("Done testing on treebank");
    // ruleErrorEval.display(true, pwErr);
    return pcfgPE.getEvalbF1();
  }


  private static void printArgs(String[] args, PrintStream ps) {
    ps.print("LexicalizedParser invoked with arguments:");
    for (int i = 0; i < args.length; i++) {
      ps.print(" " + args[i]);
    }
    ps.println();
  }


  /** A simple main program for using the parser.
   *  This provides three modes of usage: one for building and serializing
   *  a parser from treebank data, one for parsing sentences from a file
   *  or URL containing a serialized or text grammar parser,
   *  and a third (mainly for parser quality testing)
   *  for training and testing a parser all in one go. <p>
   *  Usages: <br><code>
   *  java -mx1500m edu.stanford.nlp.parser.lexparser.LexicalizedParser [-v] 
   *      -train trainFilesPath start stop serializedGrammarFilename<p>
   *  java -mx1500m edu.stanford.nlp.parser.lexparser.LexicalizedParser [-v] -validate
   *      trainFilesPath start stop
   *      -treebank testFilePath start stop<p>
   *  java -mx512m edu.stanford.nlp.parser.lexparser.LexicalizedParser
   *      [-v] serializedGrammarPath filename+<p>
   *  java -mx512m edu.stanford.nlp.parser.lexparser.LexicalizedParser
   *      [-v] serializedGrammarPath -treebank testFilePath start stop
   *  </code><p>
   *  If the <code>serializedGrammarPath</code> ends in <code>.gz</code>,
   *  then the grammar is written and read as a compressed file (GZip).
   *  If the <code>serializedGrammarPath</code> is a URL, starting with
   *  <code>http://</code>, then the parser is read from the URL.
   *  By default the parser will be written as a serialized Java object file;
   *  if desired, the file format can be specified with the following alternate usage:<p>
   * <code>java edu.stanford.nlp.parser.lexparser.LexicalizedParser [-v] -train
   *      trainFilesPath [start stop] [-saveToSerializedFile grammarPath |
   *      -saveToTextFile grammarPath]</code><p>
   *  If no files are supplied in the third usage, then a hardwired sentence
   *  is parsed.
   *  All final arguments are passed to FactoredParser.<p>
   *  In the same position as the verbose flag (<code>-v</code>), many other
   *  options can be specified.  The most useful to an end user are:
   *  <UL>
   *  <LI><code>-tLPP class</code> Specify a different
   *     TreebankLangParserParams, for when using a different language or
   *     treebank (the default is English Penn Treebank)</LI>
   *  <LI><code>-encoding charset</code> Specify the character encoding of the
   *     input files</LI>
   *  <LI><code>-tokenized</code> Says that the input is already separated
   *     into whitespace-delimited tokens</LI>
   *  <LI><code>-tokenizerFactory class</code> Specifies a
   *     TokenizFromReaderFactory class to be used for tokenization</LI>
   *  <LI><code>-parseInside regexp</code> Specifies that parsing should only
   *     be done for tokens inside elements of the indicated XML-style
   *     elements (done as simple pattern matching, rather than XML parsing).
   *     For example, if this is specified as <code>text|doc</code>, then
   *     the material in <code>text</code> and <code>doc</code> elements
   *     would be parsed.  Sentences cannot span elements.
   *  <LI><code>-sentences token </code> Specifies a token that marks sentence
   *     boundaries.  Most tokens are interpreted literally, and must be
   *     tokens returned by the tokenizer.  A token starting with "&lt;" and
   *     ending with "&gt;" will cause the tokenizer to divide sentences on
   *     either this literal tag or the corresponding end tag, and a value of
   *     "newline" causes sentence breaking on newlines.</LI>
   *  <LI><code>-tagSeparator char</code> Specifies to look for tags on words
   *     separated by a reserved character char.</LI>
   *  <LI><code>-maxLength leng</code> Specify the longest sentence that
   *     will be parsed (and hence indirectly the amount of memory 
   *     needed).</LI>
   *  <LI><code>-outputTreeFormat style</code> Choose the style of output
   *     sentences: <code>penn</code> for prettyprinting as in the Penn
   *     treebank files, or <code>oneline</code> for printing sentences one
   *     per line.</LI>
   *  </UL>
   *  See also the package documentation for more details and examples of use.
   *
   *  @param args Command line arguments, as above
   */
  public static void main(String[] args) {
    boolean train = false;
    boolean saveToSerializedFile = false;
    boolean saveToTextFile = false;
    TokenizerFactory tokenizerFactory = null;
    boolean tokenized = false;
    char tagSeparator = 0;
    boolean sentencesSplit = false;
    String sentenceBoundary = null;
    Pattern parseInsideBeginPattern = null;
    Pattern parseInsideEndPattern = null;
    String suffix = "mrg";
    String serializedInputFileOrUrl = null;
    String textInputFileOrUrl = null;
    String serializedOutputFileOrUrl = null;
    String textOutputFileOrUrl = null;
    String treebankPath = null;
    Treebank testTreebank = null;
    boolean testGrammarCoverage = false;
    int low = -1;
    int high = -1;
    int argIndex = 0;
    if (args.length < 1) {
      System.err.println("usage: " +
                 "java edu.stanford.nlp.parser.lexparser.LexicalizedParser " +
                 "parserFileOrUrl filename*");
      System.exit(1);
    }

    // while loop through option arguments
    while (argIndex < args.length && args[argIndex].startsWith("-")) {
      if (args[argIndex].equalsIgnoreCase("-train")) {
        train = true;
        saveToSerializedFile = true;
        int numSubArgs = numSubArgs(args, argIndex);
        argIndex++;
        if (numSubArgs >= 1) {
          treebankPath = args[argIndex];
          argIndex++;
        }
        if (numSubArgs >= 3) {
          low = Integer.parseInt(args[argIndex]);
          high = Integer.parseInt(args[argIndex + 1]);
          argIndex += 2;
        }
      } else if (args[argIndex].equalsIgnoreCase("-tLPP") && (argIndex + 1 < args.length)) {
        try {
          Options.get().tlpParams = (TreebankLangParserParams)
                  Class.forName(args[argIndex + 1]).newInstance();
        } catch (Exception e) {
          System.err.println("Couldn't instantiate TreebankLangParserParams "
                             + args[argIndex + 1]);
        }
        argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-encoding")) { // sets encoding for TreebankLangParserParams

        Options.get().tlpParams.setInputEncoding(args[argIndex + 1]);
        Options.get().tlpParams.setOutputEncoding(args[argIndex + 1]);
        argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-validate")) {
        train = true;
        argIndex++;
      } else if (args[argIndex].equalsIgnoreCase("-suffix")) {
        suffix = args[argIndex + 1];
        argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-tokenized")) {
        tokenized = true;
        argIndex += 1;
      } else if (args[argIndex].equalsIgnoreCase("-tokenizerFactory")) {
        try {
          tokenizerFactory = (TokenizerFactory)
                  Class.forName(args[argIndex + 1]).newInstance();
        } catch (Exception e) {
          System.err.println("Couldn't instantiate TokenizerFactory " +
                             args[argIndex + 1]);
        }
        argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-sentences")) {
        sentencesSplit = true;
        sentenceBoundary = args[argIndex + 1];
        argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-parseInside")) {
        try {
          String beginRegex = "<(?:" + args[argIndex + 1] + ")>";
          String endRegex = "</(?:" + args[argIndex + 1] + ")>";
          parseInsideBeginPattern = Pattern.compile(beginRegex);
          parseInsideEndPattern = Pattern.compile(endRegex);
          sentencesSplit = true;
          if (sentenceBoundary == null) {
            sentenceBoundary = "";
          }
        } catch (PatternSyntaxException pse) {
          System.err.println("Bad parseInside regular expression: " + pse);
        }
        argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-tagSeparator")) {
        tagSeparator = args[argIndex + 1].charAt(0);
        argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-loadFromSerializedFile")) {
        // load the parser from a binary serialized file
        // the next argument must be the path to the parser file
        serializedInputFileOrUrl = args[argIndex + 1];
        argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-loadFromTextFile")) {
        // load the parser from declarative text file
        // the next argument must be the path to the parser file
        textInputFileOrUrl = args[argIndex + 1];
        argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-saveToSerializedFile")) {
        saveToSerializedFile = true;
        serializedOutputFileOrUrl = args[argIndex + 1];
        argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-saveToTextFile")) {
        // save the parser to declarative text file
        saveToTextFile = true;
        int numSubArgs = numSubArgs(args, argIndex);
        textOutputFileOrUrl = args[argIndex + 1];
        argIndex += 2;
      } else if (args[argIndex].equalsIgnoreCase("-treebank")) {
        // the next argument is the treebank path and range for testing
        testTreebank = Options.get().tlpParams.testMemoryTreebank();
        int testlow = Integer.parseInt(args[argIndex + 2]);
        int testhigh = Integer.parseInt(args[argIndex + 3]);
        testTreebank.loadPath(args[argIndex + 1],
                              new NumberRangeFileFilter(testlow, testhigh, true));
        argIndex += 4;
      } else if (args[argIndex].equalsIgnoreCase("-testGrammarCoverage")) {
        // the next argument is the treebank path and range for testing
        testGrammarCoverage = true;
        testTreebank = Options.get().tlpParams.testMemoryTreebank();
        int testlow = Integer.parseInt(args[argIndex + 2]);
        int testhigh = Integer.parseInt(args[argIndex + 3]);
        testTreebank.loadPath(args[argIndex + 1],
                              new NumberRangeFileFilter(testlow, testhigh, true));
        argIndex += 4;
      } else {
        int j = FactoredParser.setOptionFlag(args, argIndex);
        if (j == argIndex) {
          j = Options.get().tlpParams.setOptionFlag(args, argIndex);
        }
        if (j == argIndex) {
          System.err.println("Unknown option ignored: " + args[argIndex]);
          j++;
        }
        argIndex = j;
      }
    } // end while loop through arguments

    // all other arguments are order dependent and
    // are processed in order below

    TreePrint treePrint = Test.treePrint();
    treePrint.setHeadFinder(Options.get().tlpParams.headFinder());

    // YOU HAVE TO DO THIS CRAP
    // TODO: FIX THAT
    Train.splitters = new HashSet(Arrays.asList(Options.get().tlpParams.splitters()));
    Train.sisterSplitters = new HashSet(Arrays.asList(Options.get().tlpParams.sisterSplitters()));
    BinarizerFactory.TreeAnnotator.setTreebankLang(Options.get().tlpParams);

    LexicalizedParser lp = null;
    if ( ! train && Test.verbose) {
      System.out.println("Currently " + new Date());
      printArgs(args, System.out);
    }
    if (train) {
      printArgs(args, System.out);
      // so we train a parser using the treebank
      if (treebankPath == null) {
        // the next arg must be the treebank path, since it wasn't give earlier
        treebankPath = args[argIndex];
        argIndex++;
        if (args.length > argIndex + 1) {
          // the next two args might be the range
          low = Integer.parseInt(args[argIndex]);
          high = Integer.parseInt(args[argIndex + 1]);
          argIndex += 2;
        }
      }
      GrammarCompactor compactor = null;
      if (Train.compactGrammar() == 3) {
        compactor = new ExactGrammarCompactor(false, false);
      }
      if (low >= 0 && high >= 0) {
        // we were given a range
        lp = new LexicalizedParser(treebankPath,
                                   new NumberRangeFileFilter(low, high, true),
                                   Options.get().tlpParams, compactor);
      } else {
        lp = new LexicalizedParser(treebankPath, Options.get().tlpParams,
				   compactor);
      }
    } else if (textInputFileOrUrl != null) {
      // so we load the parser from a text grammar file
      lp = new LexicalizedParser(textInputFileOrUrl, true);
    } else {
      // so we load a serialized parser
      if (serializedInputFileOrUrl == null) {
        // the next argument must be the path to the serialized parser
        serializedInputFileOrUrl = args[argIndex];
        argIndex++;
      }
      try {
        lp = new LexicalizedParser(serializedInputFileOrUrl);
      } catch (IllegalArgumentException e) {
        System.err.println("Error loading parser, exiting...");
        System.exit(0);
      }
    }

    // at this point we should be sure that Options.get().tlpParams is
    // set appropriately (from command line, or from grammar file),
    // and will never change again.  We also set the tlpParams of the
    // LexicalizedParser instance to be the same object.  This is
    // redundancy that we probably should take out eventually.
    //
    // -- Roger
    if (Test.verbose) {
	System.err.println("ParserPack is " + 
			   Options.get().tlpParams.getClass().getName());
	System.err.println("Lexicon is " + lp.pd.lex.getClass().getName());
    }

    lp.setTreebankLangParserParams(Options.get().tlpParams);
    PrintWriter pwOut = Options.get().tlpParams.pw();
    PrintWriter pwErr = Options.get().tlpParams.pw(System.err);


    // Now what do we do with the parser we've made
    if (saveToTextFile) {
      // save the parser to textGrammar format
      if (textOutputFileOrUrl != null) {
        saveParserDataToText(lp.pd, textOutputFileOrUrl);
      } else {
        System.err.println("Usage: must specify a text grammar output path");
      }
    }
    if (saveToSerializedFile) {
      if (serializedOutputFileOrUrl == null && argIndex < args.length) {
        // the next argument must be the path to serialize to
        serializedOutputFileOrUrl = args[argIndex];
        argIndex++;
      }
      if (serializedOutputFileOrUrl != null) {
        saveParserDataToSerialized(lp.pd, serializedOutputFileOrUrl);
      } else if (textOutputFileOrUrl == null) {
	// no saving/parsing request has been specified
        System.err.println("usage: " +
                "java edu.stanford.nlp.parser.lexparser.LexicalizedParser " +
                "-train trainFilesPath [start stop] serializedParserFilename");
      }
    }
    if (Test.verbose) {
      printOptions(false);
    }
    if (testTreebank != null || (argIndex < args.length && args[argIndex].equalsIgnoreCase("-treebank"))) {
      // test parser on treebank
      if (testTreebank == null) {
        // the next argument is the treebank path and range for testing
        testTreebank = Options.get().tlpParams.testMemoryTreebank();
        if (args.length < argIndex + 4) {
          testTreebank.loadPath(args[argIndex + 1]);
        } else {
          int testlow = Integer.parseInt(args[argIndex + 2]);
          int testhigh = Integer.parseInt(args[argIndex + 3]);
          testTreebank.loadPath(args[argIndex + 1],
                                new NumberRangeFileFilter(testlow, testhigh, true));
        }
      }
      if (testGrammarCoverage) {
        lp.testGrammarCoverage(testTreebank);
      } else {
        lp.testOnTreebank(testTreebank);
      }
    } else if (argIndex >= args.length) {
      // no more arguments, so we just parse our own test sentence
      String[] sentence = {"This", "is", "just", "a", "test", "."};
      // String[] sentence = { "", "", "ѧУ", "", "ѧϰ", ""};
      lp.parse(Arrays.asList(sentence));
      treePrint.printTree(lp.getBestParse(), pwOut);
    } else {
      // the rest of the arguments must be names of files to parse
      // so we parse through the various files left on the command line
      // System.err.println("Parsing files: ");
      // choose a tokenizer
      if (tokenizerFactory == null) {
        if (!tokenized) {
          // no Tokenizer was specified on the command line
          // so let's see if there is a recommended one in the language pack
          tokenizerFactory = Options.get().langpack().getTokenizerFactory();
          if (tokenizerFactory == null) {
            // none was specified in the language pack, so we use PTBTokenizer as default
            tokenizerFactory = PTBTokenizer.factory(sentencesSplit && sentenceBoundary.equalsIgnoreCase("newline"));
          }
        } else {
          // it was already tokenized, so we just split on whitespace
          tokenizerFactory = SimpleTokenizer.factory(sentencesSplit && sentenceBoundary.equalsIgnoreCase("newline"));
        }
      }
      if (Test.verbose) {
	System.err.println("TokenizerFactory is " + 
			   tokenizerFactory.getClass().getName());
      }
      // choose a sentence splitter
      WordToSentenceProcessor wordToSent;
      if (sentencesSplit) {
        Set s;
        if (sentenceBoundary.length() > 0) {
          int sbl = sentenceBoundary.length();
          if (sentenceBoundary.charAt(0) == '<' &&
                  sentenceBoundary.charAt(sbl - 1) == '>') {
            String beg = "<" + sentenceBoundary.substring(1, sbl - 1) + ">";
            String end = "</" + sentenceBoundary.substring(1, sbl - 1) + ">";
            s = new HashSet(Arrays.asList(new String[]{beg, end}));
          } else {
            if (sentenceBoundary.equalsIgnoreCase("newline")) {
              sentenceBoundary = "\n";
            }
            s = Collections.singleton(sentenceBoundary);
          }
        } else {
          // s = Collections.EMPTY_SET;
	  s = new HashSet(Arrays.asList(Options.get().tlpParams.treebankLanguagePack().sentenceFinalPunctuationWords()));
        }
        wordToSent = new WordToSentenceProcessor(Collections.EMPTY_SET,
                                                 Collections.EMPTY_SET,
                                                 s,
                                                 parseInsideBeginPattern,
                                                 parseInsideEndPattern);
      } else {
        wordToSent = new WordToSentenceProcessor(new HashSet(Arrays.asList(
                Options.get().tlpParams.treebankLanguagePack().sentenceFinalPunctuationWords())));
      }

      // go through the documents, process them and parse them
      Document doc;
      List sentList;
      Processor tagProcessor = new WordToTaggedWordProcessor(tagSeparator);
      PTBEscapingProcessor escProcessor = new PTBEscapingProcessor();
      for (int i = argIndex; i < args.length; i++) {
        String filename = args[i];
        System.err.println("Parsing file: " + filename);
        doc = null;
        try {
          if (filename.startsWith("http://")) {
            doc = new BasicDocument(tokenizerFactory).init(new URL(filename));
            Processor notags = new StripTagsProcessor();
            doc = notags.process(doc);
          } else {
            // cdm feb 2004: this case seemed to me unneeded and wrong!
            //if (sentencesSplit && !tokenized) {
            //  doc = new BasicDocument(SimpleTokenizer.factory(sentenceBoundary.equalsIgnoreCase("newline"))).init(new InputStreamReader(new FileInputStream(filename), Options.get().langpack().getEncoding()));
            //  // we'll have to do the real tokenization later, once we have sentences
            //} else {
            doc = new BasicDocument(tokenizerFactory).init(new InputStreamReader(new FileInputStream(filename), Options.get().langpack().getEncoding()));
            // System.err.println("doc size: " + doc.size());
            //}
          }
        } catch (Exception e) {
          e.printStackTrace();
        }
        if (doc != null) {
          if (sentencesSplit && !tokenized) {
            // don't need a tagProcessor, b/c it doesn't make any sense to be tagged but not tokenized
            sentList = wordToSent.process(doc);
            // we have to retokenize the individual sentences
            SentenceRetokenizingProcessor retokprocessor = new SentenceRetokenizingProcessor(tokenizerFactory);
            sentList = retokprocessor.process(sentList);
          } else {
            doc = tagProcessor.process(doc); // word to tagged word - splitting tags just in case
            sentList = wordToSent.process(doc);
          }
          int num = 0;
          for (Iterator it = sentList.iterator(); it.hasNext();) {
	    num++;
            Sentence s = (Sentence) it.next();
            if (!(tokenizerFactory.getTokenizer(null) instanceof PTBTokenizer)) {
              s = new Sentence(escProcessor.process((List) s)); // performs the PTB escaping if necessary
            }
            pwErr.println("Parsing [sent. " + num + " len. " + s.length() + 
			  "]: " + s);
            try {
              lp.parse(s);
              // System.out.println("Score: " + lp.pparser.bestScore);
              lp.getBestParse().pennPrint(pwOut);
            } catch (UnsupportedOperationException uoe) {
              pwErr.println("Whoops, sentence too long.");
            }
          }
        }
      }
    }
  } // end main

} // end class LexicalizedParser
