package edu.stanford.nlp.parser.lexparser;

import edu.stanford.nlp.trees.TreebankLanguagePack;
import edu.stanford.nlp.util.Numberer;

import java.io.*;

public class Options implements Serializable {
  /* THESE OPTIONS MUST BE THE SAME AT TRAIN AND TEST TIME */

  private Options() {
  }  // not instantiable

  private static Options instance = new Options();

  public static Options get() {
    return instance;
  }

  public static void set(Options pt) {
    instance = pt;
  };

  public int numWords = -1;
  public int numTags = -1;
  public int numStates = -1;

  /** Use suffix and capitalization information for unknowns.
   *  0 means a single unknown token.  1 uses suffix, and capitalization.
   *  2 uses a variant (richer) form of signature.  Good.
   *  Use this one.  Using the richer signatures in versions 3 or 4 seems
   *  to have very marginal or no positive value.
   *  3 uses a richer form of signature that mimics the NER word type
   *  patterns.  4 is a variant of 2.  5 is another with more English
   *  specific morphology (good for English unknowns!)
   */
  public int useUnknownWordSignatures = 0;

  /** Words more common than this are tagged with MLE P(t|w). Default 100. The
   *  smoothing is sufficiently slight that changing this has little effect.
   */
  public int SMOOTH_IN_UNKNOWNS_THRESHOLD = 100;

  /** The treebank-specific parser parameters  to use.  */
  public TreebankLangParserParams tlpParams = 
    new EnglishTreebankParserParams();

  /** returns the treebank language pack for the treebank the parser
   * is trained on. */
  public TreebankLanguagePack langpack() {
    return tlpParams.treebankLanguagePack();
  }


  /** Smarter smoothing for rare words. */
  public boolean smartMutation = false;

  /** Forces parsing with strictly CNF grammar -- unary chains are converted
   *  to XP&YP symbols and back 
   */
  public boolean forceCNF = false;

  /** Do a PCFG parse of the sentence.  If both variables are on,
   *  also do a combined parse of the sentence.
   */
  public boolean doPCFG = true;

  /** Do a dependency parse of the sentence. */
  public boolean doDep = true;

  /** if true, any child can be the head (seems rather bad!) */
  public boolean freeDependencies = false;

  /** Whether dependency grammar considers left/right direction. Good. */
  public boolean directional = true;
  public boolean genStop = true;

  /** Use distance bins in the dependency calculations */
  public boolean distance = true;
  /** Use coarser distance (4 bins) in dependency calculations */
  public boolean coarseDistance = false;

  /* Setting this option to true means that the parser generates ALL tags in
   * the lexicon for every word (some may still be scored at -Inf, but lots
   * of tags that it's never been seen with will have positive probability).
   * If false, it only generates seen tags, and if the word was of low
   * enough frequency in the training data, other tags deemed as open class.
   */
  public boolean flexiTag = false;

  /** "double count" tags rewrites as word in PCFG and Dep parser.  Good for
   *  combined parsing only (it used to not kick in for PCFG parsing).  This
   *  option is only used at Test time, but it is now in Options, so the 
   *  correct choice for a grammar is recorded by a serialized parser.
   */
  public boolean dcTags = true;


  public void display() {
    try {
      System.err.println("Options parameters:");
      writeData(new PrintWriter(System.err));
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  public void writeData(Writer w) throws IOException {
    numWords = Numberer.getGlobalNumberer("words").total();
    numTags = Numberer.getGlobalNumberer("tags").total();
    numStates = Numberer.getGlobalNumberer("states").total();
    PrintWriter out = new PrintWriter(w);
    String str =
            "numWords=" + numWords + "\n" +
            "numTags=" + numTags + "\n" +
            "numStates=" + numStates + "\n" +
            "useUnknownWordSignatures=" + useUnknownWordSignatures + "\n" +
            "SMOOTH_IN_UNKNOWNS_THRESHOLD=" + SMOOTH_IN_UNKNOWNS_THRESHOLD + "\n" +
            "parserParams=" + tlpParams.getClass().getName() + "\n" +
            "smartMutation=" + smartMutation + "\n" +
            "doPCFG=" + doPCFG + "\n" +
            "doDep=" + doDep + "\n" +
            "freeDependencies=" + freeDependencies + "\n" +
            "directional=" + directional + "\n" +
            "genStop=" + genStop + "\n" +
            "distance=" + distance + "\n" +
            "coarseDistance=" + coarseDistance + "\n" +
            "flexiTag=" + flexiTag + "\n" +
            "dcTags=" + dcTags + "\n";
    out.print(str);
    out.flush();
  }


  /**
   * Populates data in this Options from the character stream
   * given by the Reader r.
   */
  public void readData(BufferedReader in) throws IOException {
    String line, value;
    final String TRUE = "true";
    line = in.readLine();
    value = line.substring(line.indexOf('=') + 1);
    numWords = Integer.parseInt(value);
    line = in.readLine();
    value = line.substring(line.indexOf('=') + 1);
    numTags = Integer.parseInt(value);
    line = in.readLine();
    value = line.substring(line.indexOf('=') + 1);
    numStates = Integer.parseInt(value);
    line = in.readLine();
    value = line.substring(line.indexOf('=') + 1);
    useUnknownWordSignatures = Integer.parseInt(value);
    line = in.readLine();
    value = line.substring(line.indexOf('=') + 1);
    SMOOTH_IN_UNKNOWNS_THRESHOLD = Integer.parseInt(value);
    line = in.readLine();
    value = line.substring(line.indexOf('=') + 1);
    try {
      tlpParams = (TreebankLangParserParams)Class.forName(value).newInstance();
    } catch (Exception e) {
      e.printStackTrace();
    }
    line = in.readLine();
    value = line.substring(line.indexOf('=') + 1);
    smartMutation = value.equals(TRUE);
    line = in.readLine();
    value = line.substring(line.indexOf('=') + 1);
    doPCFG = value.equals(TRUE);
    line = in.readLine();
    value = line.substring(line.indexOf('=') + 1);
    doDep = value.equals(TRUE);
    line = in.readLine();
    value = line.substring(line.indexOf('=') + 1);
    freeDependencies = value.equals(TRUE);
    line = in.readLine();
    value = line.substring(line.indexOf('=') + 1);
    directional = value.equals(TRUE);
    line = in.readLine();
    value = line.substring(line.indexOf('=') + 1);
    genStop = value.equals(TRUE);
    line = in.readLine();
    value = line.substring(line.indexOf('=') + 1);
    distance = value.equals(TRUE);
    line = in.readLine();
    value = line.substring(line.indexOf('=') + 1);
    coarseDistance = value.equals(TRUE);
    line = in.readLine();
    value = line.substring(line.indexOf('=') + 1);
    flexiTag = value.equals(TRUE);
    line = in.readLine();
    value = line.substring(line.indexOf('=') + 1);
    dcTags = value.equals(TRUE);
    line = in.readLine();
  }

  private static final long serialVersionUID = 2L;

} // end class Options

