package edu.stanford.nlp.parser.lexparser;

import java.io.*;
import java.util.*;

import edu.stanford.nlp.process.*;
import edu.stanford.nlp.trees.*;
import edu.stanford.nlp.util.*;

/** A class for customizing the print method for a
 *  <code>edu.stanford.nlp.trees.Tree</code> as the output of the
 *  parser.  
 *
 * @author Roger Levy
 */  

public class TreePrint {
  /* legal output tree formats */
  public static final String[] outputTreeFormats = { "penn","oneline","none" };
  public static final String headMark = "*H";

  private String format;
  private boolean headMarkNodes = false;
  private boolean lexicalize = false;
  private boolean printDepsSet = false;

  private HeadFinder hf;

  private PrintWriter pw = new PrintWriter(System.out,true);


  public TreePrint(String format) {
    this(format,new CollinsHeadFinder());
  }

  public TreePrint(String format,HeadFinder hf) {
    if (! Arrays.asList(outputTreeFormats).contains(format)) {
      this.format = "penn";
      throw new RuntimeException("Error: output tree format " + format + " not supported");
    }
    else
      this.format = format;

    setHeadFinder(hf);
  }




  /** prints the tree to the default PrintWriter */
  public void printTree(Tree t) {
    printTree(t,pw);
  }

  /** prints the tree. */
  public void printTree(Tree t, PrintWriter pw) {
    Tree outputTree = t;


    if(lexicalize) {
      Function a = TreeJugglers.getStringLabelTreeToCategoryWordTagTreeAppliable();
      outputTree = (Tree) a.apply(outputTree);
      outputTree.percolateHeads(hf);
    }

    if(headMarkNodes) {
      outputTree = headMarkNodes(outputTree);
    }

    if(!lexicalize) { // delexicalize the output tree
      Function a = TreeJugglers.getLabelTreeToStringLabelTreeAppliable();
      outputTree = (Tree) a.apply(outputTree);
    }


    if(format.equals("penn"))
      outputTree.pennPrint(pw);
    else if(format.equals("oneline"))
      pw.println(outputTree.toString());

    if(printDepsSet) {
      Set depsSet = new DependencyEval("").makeObjects(outputTree);
      pw.println(depsSet.toString());
    }

  }



  /** sets whether or not to lexicalize the output format of the
   * tree.*/
  public void setLexicalize(boolean b) {
    lexicalize = b;
  }


  /** sets whether or not to mark the head daughter of each non-leaf
   * node */
  public void setHeadMarkNodes(boolean b) {
    headMarkNodes = b;
  }

  /** Set whether or not to print the dependency set for the tree */
  public void setPrintDepsSet(boolean b) {
    printDepsSet = b;
  }

  /** Set the headfinder to be used for lexicalizing trees.  Relevant
      if the output is to be lexicalized.  By default,
      <code>CollinsHeadFinder</code> is used. */
  public void setHeadFinder(HeadFinder hf) { 
    this.hf  = hf; 
  }

  public HeadFinder getHeadFinder() { 
    return hf;
  }


  /** Sets the default print writer for printing trees with the instance */
  public void setPrintWriter(PrintWriter pw) {
    this.pw = pw;
  }


  public Tree headMarkNodes(Tree t) {
    return headMarkNodes(t, null);
  }

  private Tree headMarkNodes(Tree t, Tree head) {
    if(t.isLeaf())
      return t; // don't worry about head-marking leaves
    Label newLabel;
    if(t==head)
      newLabel = headMark(t.label());
    else
      newLabel = t.label();
    Tree newHead = hf.determineHead(t);
    return t.treeFactory().newTreeNode(newLabel,Arrays.asList(headMarkChildren(t,newHead)));
  }

  private Label headMark(Label l) {
    Label l1 = l.labelFactory().newLabel(l);
    l1.setValue(l1.value()+"*H");
    return l1;
  }

  private Tree[] headMarkChildren(Tree t, Tree head) {
    Tree[] kids = t.children();
    Tree[] newKids = new Tree[kids.length];
    for(int i = 0, n = kids.length; i<n; i++) {
      newKids[i] = headMarkNodes(kids[i],head);
    }
    return newKids;
  }

}
