package edu.stanford.nlp.parser.lexparser;

import edu.stanford.nlp.trees.*;
import edu.stanford.nlp.io.NumberRangeFileFilter;
import edu.stanford.nlp.util.Pair;

import java.util.*;
import java.io.Reader;

/** Class for getting an annotated treebank.
 *
 *  @author Dan Klein
 */
public class TreebankAnnotator {

  TreeTransformer treeTransformer;
  TreeTransformer treeUnTransformer;

  public List annotateTrees(List trees) {
    List annotatedTrees = new ArrayList();
    for (Iterator treeI = trees.iterator(); treeI.hasNext();) {
      Tree tree = (Tree) treeI.next();
      annotatedTrees.add(treeTransformer.transformTree(tree, null));
    }
    return annotatedTrees;
  }

  public List deannotateTrees(List trees) {
    List deannotatedTrees = new ArrayList();
    for (Iterator treeI = trees.iterator(); treeI.hasNext();) {
      Tree tree = (Tree) treeI.next();
      deannotatedTrees.add(treeUnTransformer.transformTree(tree, null));
    }
    return deannotatedTrees;
  }


  public static Pair extractGrammars(List trees) {
    BinaryGrammarExtractor binaryGrammarExtractor = new BinaryGrammarExtractor();
    Pair pair = (Pair) binaryGrammarExtractor.extract(trees);
    return pair;
  }

  public static Lexicon extractLexicon(List trees) {
    Lexicon lexicon = new Lexicon();
    lexicon.train(trees);
    return lexicon;
  }

  public static List getTrees(String path, int low, int high, int minLength, int maxLength) {
    Treebank treebank = new DiskTreebank(new TreeReaderFactory() {
      public TreeReader newTreeReader(Reader in) {
        return new PennTreeReader(in,
                                  new LabeledScoredTreeFactory(new WordFactory()),
                                  new BobChrisTreeNormalizer());
      }
    });
    treebank.loadPath(path, new NumberRangeFileFilter(low, high, true));
    List trees = new ArrayList();
    for (Iterator treeI = treebank.iterator(); treeI.hasNext();) {
      Tree tree = (Tree) treeI.next();
      if (tree.yield().size() <= maxLength &&
          tree.yield().size() >= minLength) {
        trees.add(tree);
      }
    }
    return trees;
  }

  public static List removeDependencyRoots(List trees) {
    List prunedTrees = new ArrayList();
    for (Iterator treeI = trees.iterator(); treeI.hasNext();) {
      Tree tree = (Tree) treeI.next();
      prunedTrees.add(removeDependencyRoot(tree));
    }
    return prunedTrees;
  }

  static Tree removeDependencyRoot(Tree tree) {
    List childList = tree.getChildrenAsList();
    Tree last = (Tree) childList.get(childList.size()-1);
    if (!last.label().value().equals(Lexicon.BOUNDARY_TAG))
      return tree;
    List lastGoneList = childList.subList(0, childList.size()-1);
    tree.setChildren(lastGoneList);
    return tree;
  }

  public Tree collinize(Tree tree) {
    return Options.get().tlpParams.collinizer().transformTree(tree, new LabeledScoredTreeFactory());
  }

  public TreebankAnnotator() {
    Options.get().tlpParams = new EnglishTreebankParserParams();
    Train.splitters = new HashSet(Arrays.asList(Options.get().tlpParams.splitters()));
    Train.sisterSplitters = new HashSet(Arrays.asList(Options.get().tlpParams.sisterSplitters()));
    String[] args = new String[] {"-acl03pcfg", "-cnf"};
    FactoredParser.setOptionFlag(args,0);
    FactoredParser.setOptionFlag(args,1);
    treeTransformer = new BinarizerFactory.TreeAnnotatorAndBinarizer(new CollinsHeadFinder(), ! Train.outsideFactor(), true);
    BinarizerFactory.TreeAnnotator.setTreebankLang(Options.get().tlpParams);
    treeUnTransformer = BinarizerFactory.getDebinarizer();
  }


  public static void main(String[] args) {
    CategoryWordTag.printWordTag = false;
    String path = args[0];
    List trees = getTrees(path, 200, 219, 0, 10);
    ((Tree) trees.iterator().next()).pennPrint();
    List annotatedTrees = TreebankAnnotator.removeDependencyRoots(new TreebankAnnotator().annotateTrees(trees));
    ((Tree) annotatedTrees.iterator().next()).pennPrint();
  }

}
