package edu.stanford.nlp.process;

import java.util.regex.Pattern;
import java.util.HashMap;

import edu.stanford.nlp.trees.HasWord;


/** Takes a HasWord or String and returns a lowercase version of it.
 *  This is deterministic spelling coversion, and so cannot deal with
 *  certain cases involving complex ambiguities, but it can do most of the
 *  simple case of English to American conversion.
 *  <i>This list is still quite incomplete, but does some of the 
 *  commenest cases found when running our parser or doing biomedical 
 *  processing.</i>
 *
 *  @author Christopher Manning
 */
public class Americanize implements Function {

    private static boolean staticCapitalizeTimex = true;

    private boolean capitalizeTimex = true;

    public static final int CAPITALIZE_TIMEX = 1;

    public Americanize() {
    }

    public Americanize(int flags) {
	if ((flags & CAPITALIZE_TIMEX) != 0) {
	    capitalizeTimex = true;
	}
    }


    /**
     * Americanize the HasWord or String coming in.
     * @param in A HasWord or String to covert to American if needed.
     * @return Either the input or an Americanized version of it.
     */
    public Object apply(Object in) {
	if (in instanceof HasWord) {
	    HasWord w = (HasWord) in;
	    String str = w.word();
	    String outStr = americanize(str);
	    if ( ! outStr.equals(str)) {
		w.setWord(outStr);
	    }
	    return w;
	} else {
	    // assume a String
	    String str = (String) in;
	    return americanize(str);
	}
    }


    /** Convert the spelling of a word from British to American English.
     *  This is deterministic spelling coversion, and so cannot deal with
     *  certain cases involving complex ambiguities, but it can do most of the
     *  simple case of English to American conversion.
     */
    public static String americanize(String str) {
	return americanize(str, staticCapitalizeTimex);
    }


    /** Convert the spelling of a word from British to American English.
     *  This is deterministic spelling coversion, and so cannot deal with
     *  certain cases involving complex ambiguities, but it can do most of the
     *  simple case of English to American conversion.
     */
    public static String americanize(String str, boolean capitalizeTimex) {
	// System.err.println("str is |" + str + "|");
	// System.err.println("timexMapping.contains is " + 
	//            timexMapping.containsKey(str));
	if (capitalizeTimex && timexMapping.containsKey(str)) {
	    return (String) timexMapping.get(str);
	} else if (mapping.containsKey(str)) {
	    return (String) mapping.get(str);
	} else {
            for (int i = 0; i < pats.length; i++) {
                if (pats[i].matcher(str).find()) {
                    // System.err.println("Replacing " + word + " with " +
                    //             pats[i].matcher(word).replaceAll(reps[i]));
                    return pats[i].matcher(str).replaceAll(reps[i]);
                }
            }
	    return str;
	}
    }


    static Pattern[] pats = {Pattern.compile("haemo"),
                             Pattern.compile("haemato"),
                             Pattern.compile("aemia$"),
                             Pattern.compile("programme(s?)$"),
                             Pattern.compile("([A-Za-z][a-z]{2})our$"),
                             Pattern.compile("centre") };

    static String[] reps = {"hemo", "hemato", "emia", 
			    "program$1", "$1or", "center" };

    // do some normalization and British -> American mapping!
    private static final String[] converters = new String[] 
        { "anaesthetic", "analogue", "analogues",
	  "analyse", "analysed", "analysing", /* not analyses NNS */
	  "behaviour", "behaviours", "candour", 
	  "capitalise", "capitalised", "capitalisation",
	  "colour", "colours", "coloured", "colouring",
	  "defence", "dialogue", "dialogues", 
	  "discolour", "discolours", "discoloured", "discolouring",
	  "endeavour", "endeavours", "endeavoured", "endeavouring", "fervour",
	  "favour", "favours", "favoured", "favouring", 
	  "favourite", "favourites", 
	  "finalise", "finalised", "finalising",
	  "flavour", "flavours",
	  "flavoured", "flavouring", "grey", "harbour", "harbours", 
	  "homologue", "homologues",
	  "honour", "honours", "honoured", "honouring",
	  "humour", "humours", "humoured", "humouring",
	  "labour", "labours", "laboured", "labouring", 
	  "localise", "localised",
	  "maximise", "maximised", "maximising",
	  "minimise", "minimised", "minimising",
	  "modernise", "modernised", "modernising",
	  "misdemeanour", "misdemeanours", "neighbour", "neighbours", 
	  "penalise", "penalised",
	  "rumour", "rumours", "rumoured", "rumouring",
	  "savour", "savours", "savoured", "savouring",
	  "splendour", "splendours", "vapour", "vapours", "vigour" };

    private static final String[] converted = new String[]
        { "anesthetic", "analog", "analogs",
	  "analyze", "analyzed", "analyzing",
	  "behavior", "behaviors", "candor", 
	  "capitalize", "capitalized", "capitalization",
	  "color", "colors", "colored", "coloring",
	  "defense", "dialog", "dialogs", 
	  "discolor", "discolors", "discolored", "discoloring",
	  "endeavor", "endeavors", "endeavored", "endeavoring", "fervor",
	  "favor", "favors", "favored", "favoring", 
	  "favorite", "favorites", 
	  "finalize", "finalized", "finalizing",
	  "flavor", "flavors",
	  "flavored", "flavoring", "gray", "harbor", "harbors", 
	  "homolog", "homologs",
	  "honor", "honors", "honored", "honoring", 
	  "humor", "humors", "humored", "humoring",
	  "labor", "labors", "labored", "laboring", 
	  "localize", "localized",
	  "maximize", "maximized", "maximizing",
	  "minimize", "minimized", "minimizing",
	  "misdemeanor", "misdemeanors", "neighbor", "neighbors",
	  "modernize", "modernized", "modernizing",
	  "penalize", "penalized",
	  "rumor", "rumors", "rumored", "rumoring",
	  "savor", "savors", "savored", "savoring",
	  "splendor", "splendors", "vapor", "vapors", "vigor" };

    private static final String[] timexConverters = new String[] 
        { "january", "february", /* not "march" ! */
	  "april", /* Not "may"! */ "june", "july", "august", "september",
	  "october", "november", "december", "monday", "tuesday", "wednesday",
	  "thursday", "friday", "saturday", "sunday" };

    private static final String[] timexConverted = new String[]
        { "January", "February", /* not "march" ! */
	  "April", /* Not "may"! */ "June", "July", "August", "September",
	  "October", "November", "December", "Monday", "Tuesday", "Wednesday",
	  "Thursday", "Friday", "Saturday", "Sunday" };

    private static HashMap mapping = new HashMap();

    private static HashMap timexMapping = new HashMap();


    // static initialization block
    static {
	if (converters.length != converted.length ||
	        timexConverters.length != timexConverted.length ||
	        pats.length != reps.length) {
	    throw new RuntimeException("Americanize: Bad initialization data");
	}
	for (int i = 0; i < converters.length; i++) {
	    mapping.put(converters[i], converted[i]);
	}
	for (int i = 0; i < timexConverters.length; i++) {
	    timexMapping.put(timexConverters[i], timexConverted[i]);
	}
    }


    public void setStaticCapitalizeTimex(boolean capitalizeTimex) {
	staticCapitalizeTimex = capitalizeTimex;
    }


    /** Americanize and print the command line arguments.
     *  This main method is just for debugging.  
     */
    public static void main(String[] args) {
	System.err.println("capitalizeTimex is " + staticCapitalizeTimex);
	System.err.println("mapping has " + mapping.size() + " mappings.");
	System.err.println("timexMapping has " + timexMapping.size() +
			   " mappings.");
	System.err.println();
	for (int i = 0; i < args.length; i++) {
	    System.out.println(args[i] + " --> " + americanize(args[i]));
	}
    }

}
