package edu.stanford.nlp.process;


import java.io.*;
import java.util.*;

import edu.stanford.nlp.trees.Word;
import edu.stanford.nlp.objectbank.TokenizerFactory;

/**
 * A Tokenizer that splits only on white space (spaces, tabs, and
 * carriage returns).  Essentially, assumes that the input is
 * pre-tokenized.  Currently doesn't support peek() or remove(),
 * although it might in the future.
 *
 * @author Roger Levy (rog@stanford.edu)
 */
public class DummyTokenizer extends AbstractTokenizer {
  DummyLexer lexer;

  /** Constructs a new DummyTokenizer that treats carriage returns as normal whitespace. */
  public DummyTokenizer(Reader r) {
    lexer = new DummyLexer(r);
    getNext();
  }

  /**
   * Gets the next valid token from the input stream.
   */
  protected Object getNext() {
    if (lexer == null) {
      return null;
    }
    Object token = null;
    try {
      token = lexer.next();
    } catch (Exception e) {
    }
    return token;
  }

  public static TokenizerFactory factory() {
    return new DummyTokenizerFactory();
  }

  static class DummyTokenizerFactory implements TokenizerFactory {
    public DummyTokenizerFactory() {
    }
    public Iterator getIterator(Reader r) {
      return getTokenizer(r);
    }
    public Tokenizer getTokenizer(Reader r) {
      return new DummyTokenizer(r);
    }
  }

  /** Reads a file from the argument and prints its tokens one per line.
   *  This is mainly as a testing aid, but it can also be quite useful
   *  standalone to turn a corpus into a one token per line file of tokens.
   *  <p>
   *  Usage: <code>java edu.stanford.nlp.process.DummyTokenizer filename
   *  </code>
   *  @param args Command line arguments
   */
  public static void main(String[] args) throws IOException {
    if (args.length < 1) {
      System.err.println("usage: java edu.stanford.nlp.process." +
              "DummyTokenizer filename");
      return;
    }
    Tokenizer tokenizer = new DummyTokenizer(new FileReader(args[0]));
    List words = tokenizer.tokenize();
    for (int i = 0; i < words.size(); i++)
      System.out.println(words.get(i));
  }

}

