package edu.stanford.nlp.process;

import java.io.*;
import java.util.*;

import edu.stanford.nlp.io.*;


/** An implementation of {@link Tokenizer} designed to work with
 *  {@link Lexer} implementing classes.  Throw in a {@link Lexer} on
 *  construction and you get a {@link Tokenizer}.
 *
 *  @author Roger Levy
 */
public class LexerTokenizer extends AbstractTokenizer implements Tokenizer {

  private Lexer l;

  protected Object getNext() {
    Object token = null;
    try {
      int a = l.IGNORE;
      while ((a = l.yylex()) == l.IGNORE) ; // skip tokens to be ignored
      if (a == l.getYYEOF()) {
        token = null;
      } else {
        token = l.yytext();
      }
    } catch (IOException e) {
      throw new RuntimeIOException(e.getMessage());
    }
    return token;
  }

  /* Constructs a tokenizer from a {@link Lexer}
   */
  public LexerTokenizer(Lexer l) {
    if (l == null)
      throw new IllegalArgumentException("You can't make a Tokenizer out of a null Lexer!");
    this.l = l;
  }

  /* Constructs a tokenizer from a {@link Lexer} and makes a {@link
   * Reader} the active input stream for the tokenizer.
   */
  public LexerTokenizer(Lexer l, Reader r) {
    this(l);

    try {
      l.yyreset(r);
    } catch (IOException e) {
      throw new RuntimeIOException(e.getMessage());
    }

    getNext();
  }


  /** for testing only */
  public static void main(String[] args) throws IOException {
    Tokenizer t = new LexerTokenizer(new JFlexDummyLexer((Reader) null),
            new BufferedReader(new FileReader(args[0])));
    while (t.hasNext()) {
      System.out.println("token " + t.next());
    }
  }

}
