package edu.stanford.nlp.process;

import java.util.*;
import java.net.*;
import java.io.File;

import edu.stanford.nlp.dbm.*;
import edu.stanford.nlp.trees.*;

/**
 * Produces a new Document of Words in which special characters of the PTB 
 * have been properly escaped. 
 *
 * @author Teg Grenager (grenager@stanford.edu)
 */
public class PTBEscapingProcessor implements Processor {

  protected Map stringSubs;
  protected char[] oldChars;

  // these should all appear as tokens themselves
  protected static final String[] oldStrings = { "(",     ")",     "[",     "]",     "{",     "}",     "*",   "/"  };
  protected static final String[] newStrings = { "-LRB-", "-RRB-", "-LSB-", "-RSB-", "-LSB-", "-RSB-", "\\*", "\\/"};

  // these are chars that might appear inside tokens
  protected static final char[] defaultOldChars =   { '*',   '/'   };

  protected boolean fixQuotes = true;

  public PTBEscapingProcessor() {
    this(makeStringMap(), defaultOldChars, true);
  }

  public PTBEscapingProcessor(Map stringSubs, char[] oldChars, boolean fixQuotes) {
    this.stringSubs = stringSubs;
    this.oldChars = oldChars;
    this.fixQuotes = fixQuotes;
  }

  protected static Map makeStringMap() {
    Map map = new HashMap();
    for (int i=0; i<oldStrings.length; i++) {
      map.put(oldStrings[i], newStrings[i]);
    }
    return map;
  }

  public Document process(Document input) {
    Document result = input.blankDocument();
    result.addAll(process((List)input));
    return result;
  }

  /** @param input must be a List of objects of type HasWord */
  public List process(List input) {
    List output = new ArrayList();
    HasWord h;
    String s, newS;
    for (Iterator i = input.iterator(); i.hasNext(); ) {
      h = (HasWord) i.next();
      s = h.word();
      newS = (String) stringSubs.get(s);
      if (newS != null) {
	h.setWord(newS);
      } else {
	h.setWord(escapeString(s));
      }
      output.add(h);
    }
    if (fixQuotes) 
      return fixQuotes(output);
    return output;
  }

  private List fixQuotes(List input) {
    LinkedList result = new LinkedList();
    boolean begin;
    // see if there is a quote at the end
    if (((HasWord)input.get(input.size()-1)).word().equals("\"")) {
      // alternate from the end
      begin = false;
      for (int i=input.size()-1; i>=0; i--) {
	HasWord hw = (HasWord) input.get(i);
	String tok = hw.word();
	if (tok.equals("\"")) {
	  if (begin) {
	    hw.setWord("``");
	    begin = false;
	  } else {
	    hw.setWord("\'\'");
	    begin = true;
	  }
	} // otherwise leave it alone
	result.addFirst(hw);
      } // end loop
    } else {
      // alternate from the beginning
      begin = true;
      for (int i=0; i<input.size(); i++) {
	HasWord hw = (HasWord) input.get(i);
	String tok = hw.word();
	if (tok.equals("\"")) {
	  if (begin) {
	    hw.setWord("``");
	    begin = false;
	  } else {
	    hw.setWord("\'\'");
	    begin = true;
	  }
	} // otherwise leave it alone
	result.addLast(hw);
      } // end loop
    }
    return result;
  }

  private String escapeString(String s) {
    StringBuffer buff = new  StringBuffer();
    for (int i=0; i<s.length(); i++) {
      char curChar = s.charAt(i);
      if (curChar == '\\') {
	// add this and the next one
	buff.append(curChar);
	i++;
	curChar = s.charAt(i);
	buff.append(curChar);
      } else {
	// run through all the chars we need to escape
	for (int j=0; j<oldChars.length; j++) {
	  if (curChar == oldChars[j]) {
	    buff.append('\\');
	    break;
	  } 
	}
	// append the old char no matter what
	buff.append(curChar);
      }
    }
    return buff.toString();
  }

  /** This will do the escaping on an input file. Input file must already be tokenized,
   *  with tokens separated by whitespace. <br>
   *  Usage: java edu.stanford.nlp.process.PTBEscapingProcessor fileOrUrl
   *  @param args Command line argument: a file or URL
   
   */
  public static void main(String[] args) {
    if (args.length != 1) {
      System.out.println("usage: java edu.stanford.nlp.process.PTBEscapingProcessor fileOrUrl");
      System.exit(0);
    }
    String filename = args[0];
    try {
      Document d = null;
      if (filename.startsWith("http://")) {
	Document dpre = new BasicDocument(SimpleTokenizer.factory()).init(new URL(filename));
	Processor notags = new StripTagsProcessor();
	d = notags.process(dpre);
      } else {
	d = new BasicDocument(SimpleTokenizer.factory()).init(new File(filename));
      }
      Processor proc = new PTBEscapingProcessor();
      Document newD = proc.process(d);
      for (Iterator it = newD.iterator(); it.hasNext(); ) {
	HasWord word = (HasWord) it.next();
	System.out.println(word);
      }
    } catch (Exception e) {
      e.printStackTrace();
    }

  }

}
