package edu.stanford.nlp.process;

import java.io.*;
import java.util.Iterator;

import edu.stanford.nlp.trees.Word;
import edu.stanford.nlp.objectbank.TokenizerFactory;

/**
 * Simple Tokenizer implementation that wraps a StreamTokenizer.
 *
 * @author Joseph Smarr (jsmarr@stanford.edu)
 * @author Teg Grenager (grenager@stanford.edu)
 */
public class SimpleTokenizer extends AbstractTokenizer {

  /** A factory which vends SimpleTokenizers.
   *
   *  @author Christopher Manning
   */
  static class SimpleTokenizerFactory implements TokenizerFactory {
    boolean eolIsSignificant;
    public SimpleTokenizerFactory() {
      this(false);
    }
    public SimpleTokenizerFactory(boolean eolIsSignificant) {
      this.eolIsSignificant = eolIsSignificant;
    }
    public Iterator getIterator(Reader r) {
      return getTokenizer(r);
    }
    public Tokenizer getTokenizer(Reader r) {
      return new SimpleTokenizer(r, eolIsSignificant);
    }
  }

  StreamTokenizer st;
  boolean eolIsSignificant;

  protected Object getNext() {
    Object token = null;
    if (st == null) {
      return null;
    }
    try {
      st.nextToken();
      if (st.ttype == StreamTokenizer.TT_EOF) {
        return null;
      } else if (st.ttype == StreamTokenizer.TT_EOL) {
        if (eolIsSignificant) {
          token = new Word("\n");
        } else {
          return null;
        }
      } else {
        token = new Word(st.sval);
      }
    } catch (Exception e) {
    }
    return token;
  }

  /**
   * Constructs a new SimpleTokenizer with the Reader r as its source.
   */
  public SimpleTokenizer(Reader r) {
    this(r, false);
  }


  /**
   * Constructs a new SimpleTokenizer with the Reader r as its source.
   */
  public SimpleTokenizer(Reader r, boolean eolIsSignificant) {
    this.eolIsSignificant = eolIsSignificant;
    // The conditional below is currently needed in LexicalizedParser, since 
    // it passes in a null arg while doing type-checking for sentence escaping
    // but StreamTokenizer barfs on that.  But maybe shouldn't be here.
    if (r != null) {
      setSource(r);
    }
  }

  /**
   * Sets the source of this Tokenizer to be the Reader r.
   */
  private void setSource(Reader r) {
    st = new StreamTokenizer(r);
    st.resetSyntax();
    st.whitespaceChars(0, 0x20);
    st.wordChars(0x21, 0x7E);
    st.eolIsSignificant(eolIsSignificant);
  }

  public static TokenizerFactory factory() {
    return new SimpleTokenizerFactory(false);
  }

  public static TokenizerFactory factory(boolean eolIsSignificant) {
    return new SimpleTokenizerFactory(eolIsSignificant);
  }

  // for debugging
  private String type(int n) {
    if (n == StreamTokenizer.TT_EOF) return "EOF";
    if (n == StreamTokenizer.TT_EOL) return "EOL";
    if (n == StreamTokenizer.TT_NUMBER) return "NUMBER";
    if (n == StreamTokenizer.TT_WORD) return "WORD";
    return "nothing";
  }

  /** Reads a file from the argument and prints its tokens one per line.
   *  This is mainly as a testing aid, but it can also be quite useful
   *  standalone to turn a corpus into a one token per line file of tokens.
   *  <p>
   *  Usage: <code>java edu.stanford.nlp.process.SimpleTokenizer filename
   *  </code>
   *  @param args Command line arguments
   */
  public static void main(String[] args) throws IOException {
    if (args.length < 1) {
      System.err.println("usage: java edu.stanford.nlp.process." +
              "SimpleTokenizer [-cr] filename");
      return;
    }
    Tokenizer tokenizer = new SimpleTokenizer(new FileReader(args[args.length-1]), args[0].equals("-cr"));
    while (tokenizer.hasNext()) {
      System.out.println(tokenizer.next());
    }
  }


}

