package edu.stanford.nlp.util;

import java.util.*;
import java.io.*;

/**
 * A specialized kind of hash table (or map) for storing numeric counts for
 * objects. It works like a Map,
 * but with different methods for easily getting/setting/incrementing counts
 * for objects and computing various functions with the counts.
 * The Counter constructor
 * and <tt>addAll</tt> method can be used to copy another Counter's contents
 * over. This class also provides access
 * to Comparators that can be used to sort the keys or entries of this Counter
 * by the counts, in either ascending or descending order.
 *
 * @author Dan Klein (klein@cs.stanford.edu)
 * @author Joseph Smarr (jsmarr@stanford.edu)
 * @author Teg Grenager
 * @author Galen Andrew
 * @author Christopher Manning
 */
public class Counter implements Serializable {

  Map map;
  MapFactory mapFactory;

  /** Default comparator for breaking ties in argmin and argmax. */
  private static final Comparator naturalComparator = new NaturalComparator();
  private static final long serialVersionUID = 4;

  // CONSTRUCTORS

  /** Constructs a new (empty) Counter. */
  public Counter() {
    this(MapFactory.HASH_MAP_FACTORY);
  }

  /**
   * Pass in a MapFactory and the map it vends will back your counter.
   */
  public Counter(MapFactory mapFactory) {
    this.mapFactory = mapFactory;
    map = mapFactory.newMap();
  }

  /**
   * Constructs a new Counter with the contents of the given Counter.
   */
  public Counter(Counter c) {
    this();
    addAll(c);
  }


  // STANDARD ACCESS MODIFICATION METHODS

  /**
   * Returns the current total count for all objects in this Counter.
   * All counts are summed each time, so cache it if you need it repeatedly.
   */
  public double totalCount() {
    double total = 0.0;
    for (Iterator iter = map.keySet().iterator(); iter.hasNext();) {
      Object key = iter.next();
      total += getCount(key);
    }
    return total;
  }

  /**
   * Returns the total count for all objects in this Counter that pass the
   * given Filter. Passing in a filter that always returns true is equivalent
   * to calling {@link #totalCount()}.
   */
  public double totalCount(Filter filter) {
    double total = 0.0;
    for (Iterator iter = map.keySet().iterator(); iter.hasNext(); ) {
      Object key = iter.next();
      if (filter.accept(key)) total += getCount(key);
    }
    return (total);
  }

  /** Returns the mean of all the counts (totalCount/size). */
  public double averageCount() {
    return (totalCount() / map.size());
  }

  /**
   * Returns the current count for the given key, which is 0 if it hasn't
   * been
   * seen before. This is a convenient version of <code>get</code> that casts
   * and extracts the primitive value.
   */
  public double getCount(Object key) {
    Number count = (Number) map.get(key);
    if (count == null) {
      return (0.0); // haven't seen this object before -> 0 count
    }
    return (count.doubleValue());
  }

  /**
   * @deprecated Create a {@link Distribution} instead.
   */
  public double getNormalizedCount(Object key) {
    return (getCount(key) / (totalCount()));
  }

  /**
   * Sets the current count for the given key. This will wipe out any existing
   * count for that key.
   * <p>
   * To add to a count instead of replacing it, use
   * {@link #incrementCount(Object,double)}.
   */
  public void setCount(Object key, double count) {
    if (tempMDouble == null) {
      tempMDouble = new MutableDouble();
    }
    tempMDouble.set(count);
    tempMDouble = (MutableDouble) map.put(key, tempMDouble);
  }

  // for more efficient memory usage
  private transient MutableDouble tempMDouble = null;

  /**
   * Sets the current count for each of the given keys. This will wipe out
   * any existing counts for these keys.
   * <p>
   * To add to the counts of a collection of objects instead of replacing them,
   * use {@link #incrementCounts(Collection,double)}.
   */
  public void setCounts(Collection keys, double count) {
    for (Iterator iter = keys.iterator(); iter.hasNext();)
      setCount(iter.next(), count);
  }

  /**
   * Adds the given count to the current count for the given key. If the key
   * hasn't been seen before, it is assumed to have count 0, and thus this
   * method will set its count to the given amount. Negative increments are
   * equivalent to calling <tt>decrementCount</tt>.
   * <p>
   * To more conviently increment the count by 1.0, use
   * {@link #incrementCount(Object)}.
   * <p>
   * To set a count to a specifc value instead of incrementing it, use
   * {@link #setCount(Object,double)}.
   */
  public void incrementCount(Object key, double count) {
    if (tempMDouble == null)
      tempMDouble = new MutableDouble();
    tempMDouble.set(count);
    MutableDouble oldMDouble = (MutableDouble) map.put(key, tempMDouble);
    if (oldMDouble != null)
      tempMDouble.set(count + oldMDouble.doubleValue());
    tempMDouble = oldMDouble;
//      total += count;
    //put(key,new Double(count+getCount(key)));
  }

  /**
   * Adds 1.0 to the count for the given key. If the key hasn't been seen
   * before, it is assumed to have count 0, and thus this method will set
   * its count to 1.0.
   * <p>
   * To increment the count by a value other than 1.0, use
   * {@link #incrementCount(Object,double)}.
   * <p>
   * To set a count to a specifc value instead of incrementing it, use
   * {@link #setCount(Object,double)}.
   */
  public void incrementCount(Object key) {
    incrementCount(key, 1.0);
  }

  /**
   * Adds the given count to the current counts for each of the given keys.
   * If any of the keys haven't been seen before, they are assumed to have
   * count 0, and thus this method will set their counts to the given
   * amount. Negative increments are equivalent to calling <tt>decrementCounts</tt>.
   * <p>
   * To more conviniently increment the counts of a collection of objects by
   * 1.0, use {@link #incrementCounts(Collection)}.
   * <p>
   * To set the counts of a collection of objects to a specific value instead
   * of incrementing them, use {@link #setCounts(Collection,double)}.
   */
  public void incrementCounts(Collection keys, double count) {
    for (Iterator iter = keys.iterator(); iter.hasNext();)
      incrementCount(iter.next(), count);
  }

  /**
   * Adds 1.0 to the counts for each of the given keys. If any of the keys
   * haven't been seen before, they are assumed to have count 0, and thus
   * this method will set their counts to 1.0.
   * <p>
   * To increment the counts of a collection of object by a value other
   * than 1.0, use {@link #incrementCounts(Collection,double)}.
   * <p>
   * To set the counts of a collection of objects  to a specifc value instead
   * of incrementing them, use  {@link #setCounts(Collection,double)}.
   */
  public void incrementCounts(Collection keys) {
    incrementCounts(keys, 1.0);
  }

  /**
   * Subtracts the given count from the current count for the given key.
   * If the key hasn't been seen before, it is assumed to have count 0, and
   * thus this  method will set its count to the negative of the given amount.
   * Negative increments are equivalent to calling <tt>incrementCount</tt>.
   * <p>
   * To more conviently decrement the count by 1.0, use
   * {@link #decrementCount(Object)}.
   * <p>
   * To set a count to a specifc value instead of decrementing it, use
   * {@link #setCount(Object,double)}.
   */
  public void decrementCount(Object key, double count) {
    incrementCount(key, -count);
  }

  /**
   * Subtracts 1.0 from the count for the given key. If the key hasn't been
   * seen  before, it is assumed to have count 0, and thus this method will
   * set its count to -1.0.
   * <p>
   * To decrement the count by a value other than 1.0, use
   * {@link #decrementCount(Object,double)}.
   * <p>
   * To set a count to a specifc value instead of decrementing it, use
   * {@link #setCount(Object,double)}.
   */
  public void decrementCount(Object key) {
    decrementCount(key, 1.0);
  }

  /**
   * Subtracts the given count from the current counts for each of the given keys.
   * If any of the keys haven't been seen before, they are assumed to have
   * count 0, and thus this method will set their counts to the negative of the given
   * amount. Negative increments are equivalent to calling <tt>incrementCounts</tt>.
   * <p>
   * To more conviniently decrement the counts of a collection of objects by
   * 1.0, use {@link #decrementCounts(Collection)}.
   * <p>
   * To set the counts of a collection of objects to a specific value instead
   * of decrementing them, use {@link #setCounts(Collection,double)}.
   */
  public void decrementCounts(Collection keys, double count) {
    incrementCounts(keys, -count);
  }

  /**
   * Subtracts 1.0 from the counts of each of the given keys. If any of the keys
   * haven't been seen before, they are assumed to have count 0, and thus
   * this method will set their counts to -1.0.
   * <p>
   * To decrement the counts of a collection of object by a value other
   * than 1.0, use {@link #decrementCounts(Collection,double)}.
   * <p>
   * To set the counts of a collection of objects  to a specifc value instead
   * of decrementing them, use  {@link #setCounts(Collection,double)}.
   */
  public void decrementCounts(Collection keys) {
    decrementCounts(keys, 1.0);
  }

  /**
   * Adds the counts in the given Counter to the counts in this Counter.
   * <p>
   * To copy the values from another Counter rather than adding them, use
   */
  public void addAll(Counter counter) {
    for (Iterator iter = counter.keySet().iterator(); iter.hasNext();) {
      Object key = iter.next();
      double count = counter.getCount(key);
      incrementCount(key, count);
    }
  }

  /**
   * Subtracts the counts in the given Counter from the counts in this Counter.
   * <p>
   * To copy the values from another Counter rather than subtracting them, use
   */
  public void subtractAll(Counter counter) {
    for (Iterator iter = map.keySet().iterator(); iter.hasNext();) {
      Object key = iter.next();
      decrementCount(key, counter.getCount(key));
    }
  }

  // MAP LIKE OPERATIONS

  public boolean containsKey(Object key) {
    return map.containsKey(key);
  }

  /**
   * Removes the given key from this Counter. Its count will now be 0 and it
   * will no longer be considered previously seen.
   */
  public Object remove(Object key) {
//      total-=getCount(key); // subtract removed count from total (may be 0)
    return (map.remove(key));
  }

  /** Removes all the given keys from this Counter. */
  public void removeAll(Collection c) {
    for (Iterator iter = c.iterator(); iter.hasNext();)
      remove(iter.next());
  }

  /**  Removes all counts from this Counter. */
  public void clear() {
    map.clear();
//        total=0;
  }

  public int size() {
    return map.size();
  }

  public boolean isEmpty() {
    return (size()==0);
  }

  public Set keySet() {
    return map.keySet();
  }

  // OBJECT STUFF

  public boolean equals(Object o) {
    if (this == o) return true;
    if (!(o instanceof Counter)) return false;

    final Counter counter = (Counter) o;

    return map.equals(counter.map);
  }

  public int hashCode() {
    return map.hashCode();
  }

  public String toString() {
    return map.toString();
  }

  public Object clone() {
    return new Counter(this);
  }

  // EXTRA CALCULATION METHODS

  /**
   * @deprecated Create a {@link Distribution} instead.
   */
  public void normalize() {
    double total = (totalCount()); // cache value since it will change as we normalize
    for (Iterator iter = map.keySet().iterator(); iter.hasNext();) {
      Object key = iter.next();
      setCount(key, getCount(key) / total);
    }
  }

  /**
   * Removes all keys whose count is 0. After incrementing and decrementing
   * counts or adding and subtracting Counters, there may be keys left whose
   * count is 0, though normally this is undesirable. This method cleans up
   * the map.
   * <p>
   * Maybe in the future we should try to do this more on-the-fly, though it's
   * not clear whether a distinction should be made between "never seen" (i.e.
   * null count) and "seen with 0 count". Certainly there's no distinction in
   * getCount() but there is in containsKey().
   */
  public void removeZeroCounts() {
    for (Iterator iter = map.keySet().iterator(); iter.hasNext();)
      if (getCount(iter.next()) == 0)
        iter.remove();
  }

  /** Finds and returns the largest count in this Counter. */
  public double max() {
    double max = Double.NEGATIVE_INFINITY;
    for (Iterator iter = map.keySet().iterator(); iter.hasNext();)
      max = Math.max(max, getCount(iter.next()));
    return (max);
  }

  /** Finds and returns the smallest count in this Counter. */
  public double min() {
    double min = Double.POSITIVE_INFINITY;
    for (Iterator iter = map.keySet().iterator(); iter.hasNext();)
      min = Math.min(min, getCount(iter.next()));
    return (min);
  }

  /**
   * Finds and returns the key in this Counter with the largest count.
   * Ties are broken by comparing the objects using the given tie breaking
   * Comparator, favoring Objects that are sorted to the front. This is useful
   * if the keys are numeric and there is a bias to prefer smaller or larger
   * values, and can be useful in other circumstances where random tie-breaking
   * is not desirable. Returns null if this Counter is empty.
   */
  public Object argmax(Comparator tieBreaker) {
    double max = Double.NEGATIVE_INFINITY;
    Object argmax = null;
    for (Iterator iter = map.keySet().iterator(); iter.hasNext();) {
      Object key = iter.next();
      double count = getCount(key);
      if (argmax == null || count > max || (count == max && tieBreaker.compare(key, argmax) < 0)) {
        max = count;
        argmax = key;
      }
    }
    return (argmax);
  }


  /**
   * Finds and returns the key in this Counter with the largest count.
   * Ties are broken according to the natural ordering of the objects.
   * This will prefer smaller numeric keys and lexicographically earlier
   * String keys. To use a different tie-breaking Comparator, use
   * {@link #argmax(Comparator)}. Returns null if this Counter is empty.
   */
  public Object argmax() {
    return (argmax(naturalComparator));
  }

  /**
   * Finds and returns the key in this Counter with the smallest count.
   * Ties are broken by comparing the objects using the given tie breaking
   * Comparator, favoring Objects that are sorted to the front. This is useful
   * if the keys are numeric and there is a bias to prefer smaller or larger
   * values, and can be useful in other circumstances where random tie-breaking
   * is not desirable. Returns null if this Counter is empty.
   */
  public Object argmin(Comparator tieBreaker) {
    double min = Double.POSITIVE_INFINITY;
    Object argmin = null;
    for (Iterator iter = map.keySet().iterator(); iter.hasNext();) {
      Object key = iter.next();
      double count = getCount(key);
      if (argmin == null || count < min || (count == min && tieBreaker.compare(key, argmin) < 0)) {
        min = count;
        argmin = key;
      }
    }
    return (argmin);
  }

  /**
   * Finds and returns the key in this Counter with the smallest count.
   * Ties are broken according to the natural ordering of the objects.
   * This will prefer smaller numeric keys and lexicographically earlier
   * String keys. To use a different tie-breaking Comparator, use
   * {@link #argmin(Comparator)}. Returns null if this Counter is empty.
   */
  public Object argmin() {
    return (argmin(naturalComparator));
  }

  /**
   * Returns the set of keys whose counts are at or above the given threshold.
   * This set may have 0 elements but will not be null.
   */
  public Set keysAbove(double countThreshold) {
    Set keys = new HashSet();
    for (Iterator iter = map.keySet().iterator(); iter.hasNext();) {
      Object key = iter.next();
      if (getCount(key) >= countThreshold)
        keys.add(key);
    }
    return (keys);
  }

  /**
   * Returns the set of keys whose counts are at or below the given threshold.
   * This set may have 0 elements but will not be null.
   */
  public Set keysBelow(double countThreshold) {
    Set keys = new HashSet();
    for (Iterator iter = map.keySet().iterator(); iter.hasNext();) {
      Object key = iter.next();
      if (getCount(key) <= countThreshold)
        keys.add(key);
    }
    return (keys);
  }

  /**
   * Returns the set of keys that have exactly the given count.
   * This set may have 0 elements but will not be null.
   */
  public Set keysAt(double count) {
    Set keys = new HashSet();
    for (Iterator iter = map.keySet().iterator(); iter.hasNext();) {
      Object key = iter.next();
      if (getCount(key) == count)
        keys.add(key);
    }
    return (keys);
  }

  /**
   * @deprecated Use {@link Counters#entropy(Counter) } instead.
   */
  public double entropy() {
    return Counters.entropy(this);
  }

  /**
   * @deprecated Use {@link Counters#klDivergence(Counter, Counter) } instead.
   */
  public double klDivergence(Counter other) {
    return Counters.klDivergence(this, other);
  }

  /**
   * @deprecated Use {@link Counters#informationRadius(Counter, Counter) } instead.
   */
  public double informationRadius(Counter other) {
    return Counters.informationRadius(this, other);
  }

  /**
   * @deprecated Use {@link Counters#average(Counter, Counter) } instead.
   */
  public Counter average(Counter other) {
    return Counters.average(this, other);
  }

  /**
   * Returns a comparator suitable for sorting this Counter's keys or entries
   * by their respective counts. If <tt>ascending</tt> is true, lower counts
   * will be returned first, otherwise higher counts will be returned first.
   * <p>
   * Sample usage:
   * <pre>
   * Counter c = new Counter();
   * // add to the counter...
   * List biggestKeys = new ArrayList(c.keySet());
   * Collections.sort(biggestKeys, c.comparator(false));
   * List smallestEntries = new ArrayList(c.entrySet());
   * Collections.sort(smallestEntries, c.comparator(true))
   * </pre>
   */
  public Comparator comparator(boolean ascending) {
    return (new EntryValueComparator(map, ascending));
  }

  /**
   * Returns a comparator suitable for sorting this Counter's keys or entries
   * by their respective value or magnitude (unsigned value).
   * If <tt>ascending</tt> is true, smaller magnitudes will
   * be returned first, otherwise higher magnitudes will be returned first.
   * <p>
   * Sample usage:
   * <pre>
   * Counter c = new Counter();
   * // add to the counter...
   * List biggestKeys = new ArrayList(c.keySet());
   * Collections.sort(biggestKeys, c.comparator(false, true));
   * List smallestEntries = new ArrayList(c.entrySet());
   * Collections.sort(smallestEntries, c.comparator(true))
   * </pre>
   */
  public Comparator comparator(boolean ascending, boolean useMagnitude) {
    return (new EntryValueComparator(map, ascending, useMagnitude));
  }

  /**
   * Comparator that sorts objects by (increasing) count. Shortcut for calling
   * {@link #comparator(boolean) comparator(true)}.
   */
  public Comparator comparator() {
    return (comparator(true));
  }

  /** Comparator that uses natural ordering. 
   *  Returns 0 if o1 is not Comparable.
   */
  private static class NaturalComparator implements Comparator {
    public int compare(Object o1, Object o2) {
      if (o1 instanceof Comparable)
        return (((Comparable) o1).compareTo(o2));
      return (0); // soft-fail
    }
  }

  /** For internal debugging purposes only. */
  public static void main(String[] args) throws Exception {
    Counter c = new Counter();
    c.setCount("p", 0);
    c.setCount("q", 2);
    System.out.println(c + " -> " + c.totalCount() + 
		       " should be {p=0.0, q=2.0} -> 2.0");
    c.incrementCount("p");
    System.out.println(c + " -> " + c.totalCount() + 
		       " should be {p=1.0, q=2.0} -> 3.0");
    c.incrementCount("p", 2.0);
    System.out.println(c.min() + " " + c.argmin() +  
		       " should be 2.0 q");
    c.setCount("w", -5);
    c.setCount("x", -2.5);
    List biggestKeys = new ArrayList(c.keySet());
    Collections.sort(biggestKeys, c.comparator(false, true));
    System.out.println(biggestKeys + " should be w, p, x, q");
    System.out.println(c + " should be {p=3.0, q=2.0, w=-5.0, x=-2.5}");
    if (args.length > 0) {
      // serialize to  file
      ObjectOutputStream out = new ObjectOutputStream(new BufferedOutputStream(new FileOutputStream(args[0])));
      out.writeObject(c);
      out.close();

      // reconstitute
      ObjectInputStream in = new ObjectInputStream(new BufferedInputStream(new FileInputStream(args[0])));
      c = (Counter) in.readObject();
      in.close();
      System.out.println(c + " -> " + (c.totalCount()));
      //c.put("p",new Integer(3));
      //System.out.println(c+" -> "+c.total());
      System.out.println(c.min() + " " + c.argmin());
    }
  }

}
