package edu.stanford.nlp.util;

import java.util.*;
import java.io.PrintStream;

/**
 * Static methods for operating on {@link Counter}s.
 * @author Galen Andrew (galand@cs.stanford.edu)
 */
public class Counters {

  /**
   * Returns a Counter that is the union of the two Counters passed in (counts are added).
   * @param c1
   * @param c2
   * @return
   */
  public static Counter union(Counter c1, Counter c2) {
    Counter result = new Counter();
    result.addAll(c1);
    result.addAll(c2);
    return result;
  }

  /**
   * Calculates the entropy of the given counter (in bits). This method internally
   * uses normalized counts (so they sum to one), but the value returned is
   * meaningless if some of the counts are negative.
   * @return The entropy of the given counter (in bits)
   */
  public static double entropy(Counter c) {
    double entropy=0.0;
    double total = c.totalCount();
    for(Iterator iter=c.keySet().iterator();iter.hasNext();) {
      double count=c.getCount(iter.next());
      if(count==0) continue; // 0.0 doesn't add entropy but may cause -Inf
      count/=total; // use normalized count
      entropy-=count*(Math.log(count)/Math.log(2.0));
    }
    return(entropy);
  }

  /**
   * Calculates the KL divergence between the two counters.
   * That is, it calculates KL(from || to). This method internally
   * uses normalized counts (so they sum to one), but the value returned is
   * meaningless if some of the counts are negative.
   * @param from
   * @param to
   * @return The KL divergence between the distributions
   */
  public static double klDivergence(Counter from, Counter to) {
    double kl = 0.0;
    double tot = (from.totalCount());
    double tot2 = (to.totalCount());
    // System.out.println("tot is " + tot + " tot2 is " + tot2);
    for (Iterator it = from.keySet().iterator(); it.hasNext(); ) {
      Object key = it.next();
      double num = (from.getCount(key));
      if (num == 0) continue;
      num /= tot;
      double num2 = (to.getCount(key));
      num2 /= tot2;
      // System.out.println("num is " + num + " num2 is " + num2);
      kl += num * (Math.log(num/num2)/Math.log(2.0));
    }
    return kl;
  }

  /**
   * Calculates the information radius (aka the Jensen-Shannon divergence)
   * between the two Counters.  This measure is defined as:
   * <blockquote> iRad(p,q) = D(p||(p+q)/2)+D(q,(p+q)/2) </blockquote>
   * where p is one Counter, q is the other counter, and D(p||q) is the
   * KL divergence bewteen p and q.  Note that iRad(p,q) = iRad(q,p).
   * @return The information radius between the distributions
   */
  public static double informationRadius(Counter c1, Counter c2) {
    Counter avg = average(c1, c2); // (p+q)/2
    return (klDivergence(c1, avg) + klDivergence(c2, avg));
  }

  public static double cosine(Counter c1, Counter c2) {
    double dotProd = 0.0;
    double lsq1 = 0.0;
    double lsq2 = 0.0;
    for (Iterator iter1 = c1.keySet().iterator(); iter1.hasNext(); ) {
      Object key = iter1.next();
      double count1 = c1.getCount(key);
      if (count1 != 0.0) {
        lsq1 += (count1 * count1);
        double count2 = c2.getCount(key);
        if (count2 != 0.0) {
          // this is the inner product
          dotProd += (count1 * count2);
        }
      }
    }
    for (Iterator iter2 = c2.keySet().iterator(); iter2.hasNext(); ) {
      Object key = iter2.next();
      double count2 = c2.getCount(key);
      if (count2 != 0.0) {
        lsq2 += (count2 * count2);
      }
    }
    if (lsq1 != 0.0 && lsq2 != 0.0) {
      double denom = (Math.sqrt(lsq1) * Math.sqrt(lsq2));
      return dotProd / denom;
    } else {
      return 0.0;
    }
  }

  /**
   * Returns a new Counter with counts averaged from the two given Counters.
   * The average Counter will contain the union of keys in both
   * source Counters, and each count will be the average of the two source
   * counts for that key, where as usual a missing count in one Counter
   * is treated as count 0.
   * @return A new counter with counts that are the mean of the resp. counts
   * in the given counters.
   */
  public static Counter average(Counter c1, Counter c2) {
      Counter average = new Counter();
      Set allKeys = new HashSet(c1.keySet());
      allKeys.addAll(c2.keySet());
      for(Iterator iter=allKeys.iterator();iter.hasNext();) {
          Object key=iter.next();
          average.setCount(key,(c1.getCount(key) + c2.getCount(key)) * 0.5);
      }
      return(average);
  }


  /**
   * Returns a Counter which is a weighted average of c1 and c2. Counts from c1
   * are weighted with weight w1 and counts from c2 are weighted with w2.
   */
  public static Counter linearCombination(Counter c1, double w1, Counter c2, double w2) {
    Counter result = new Counter();
    for (Iterator i = c1.keySet().iterator(); i.hasNext(); ) {
      Object o = i.next();
      result.incrementCount(o, c1.getCount(o) * w1);
    }
    for (Iterator i = c2.keySet().iterator(); i.hasNext(); ) {
      Object o = i.next();
      result.incrementCount(o, c2.getCount(o) * w2);
    }
    return result;
  }

  public static Counter createCounterFromList(List l) {
    Counter result = new Counter();
    for (Iterator i = l.iterator(); i.hasNext(); ) {
      Object o = i.next();
      result.incrementCount(o);
    }
    return result;
  }

  public static List toSortedList(Counter c) {
    return toPriorityQueue(c).toSortedList();
  }

  public static PriorityQueue toPriorityQueue(Counter c) {
    PriorityQueue queue = new BinaryHeapPriorityQueue();
    for (Iterator keyI = c.keySet().iterator(); keyI.hasNext();) {
      Object key = keyI.next();
      double count = c.getCount(key);
      queue.add(key, count);
    }
    return queue;
  }

  /**
   * Great for debugging.
   * @param a
   * @param b
   */
  public static void printCounterComparison(Counter a, Counter b) {
    printCounterComparison(a, b, System.err);
  }

  /**
   * Great for debugging.
   * @param a
   * @param b
   */
  public static void printCounterComparison(Counter a, Counter b, PrintStream out) {
    if (a.equals(b)) {
      out.println("Counters are equal.");
      return;
    }
    for (Iterator keyI = a.keySet().iterator(); keyI.hasNext();) {
      Object key = (Object) keyI.next();
      double aCount = a.getCount(key);
      double bCount = b.getCount(key);
      if (Math.abs(aCount-bCount) > 1e-5)
        out.println("Counters differ on key " + key + "\t" + aCount + " vs. " + bCount);
    }
    // left overs
    Set rest = new HashSet(b.keySet());
    rest.removeAll(a.keySet());
    for (Iterator keyI = rest.iterator(); keyI.hasNext();) {
      Object key = (Object) keyI.next();
      double aCount = a.getCount(key);
      double bCount = b.getCount(key);
      if (Math.abs(aCount-bCount) > 1e-5)
        out.println("Counters differ on key " + key + "\t" + aCount + " vs. " + bCount);
    }
  }

  public static Counter getCountCounts(Counter c) {
    Counter result = new Counter();
    for (Iterator i = c.keySet().iterator(); i.hasNext();) {
      Object o = (Object) i.next();
      double count = c.getCount(o);
      result.incrementCount(new Double(count));
    }
    return result;
  }

  public static void printCounterSortedByKeys(Counter c) {
    List keyList = new ArrayList(c.keySet());
    Collections.sort(keyList);
    for (Iterator i = keyList.iterator(); i.hasNext();) {
      Object o = (Object) i.next();
      double count = c.getCount(o);
      System.out.println(o + ":" + count);
    }
  }

}
