package edu.stanford.nlp.util;

import java.io.Serializable;
import java.util.Iterator;
import java.util.Set
;

/**
 * Immutable class for representing normalized, smoothed discrete distributions
 * from {@link Counters}. Smoothed counters reserve probability mass for unseen
 * items, so queries for the probability of unseen items will return a small
 * positive amount.  {@link #totalCount} should always return 1.
 *
 * Counter passed in constructors is copied.
 *
 * @author Galen Andrew (galand@cs.stanford.edu)
 */
public class Distribution implements Serializable {

  static final long serialVersionUID = 6707148234288637809L;

  private int numberOfKeys;
  private double reservedMass;
  private Counter counter;

  public String toString() {
    return counter.toString();
  }

  public double getReservedMass() {
    return reservedMass;
  }

  public int getNumberOfKeys() {
    return numberOfKeys;
  }

  //--- cdm added Jan 2004 to help old code compile

  public Set keySet() {
    return counter.keySet();
  }

  public boolean containsKey(Object key) {
    return counter.containsKey(key);
  }

  /**
   * Returns the current count for the given key, which is 0 if it hasn't
   * been
   * seen before. This is a convenient version of <code>get</code> that casts
   * and extracts the primitive value.
   */
  public double getCount(Object key) {
    return counter.getCount(key);
  }

  //---- end cdm added


  /**
   * Creates a Distribution from the given counter, ie makes an internal
   * copy of the counter and divides all counts by the total count.
   * @param counter
   * @return a new Distribution
   */

  public static Distribution normalizedCounter(Counter counter) {
    Distribution norm = new Distribution();
    norm.counter = new Counter();
    norm.numberOfKeys = counter.size();
    norm.reservedMass = 0;
    double total = counter.totalCount();
    for (Iterator iter = counter.keySet().iterator(); iter.hasNext(); ) {
      Object key = iter.next();
      norm.counter.setCount(key, counter.getCount(key) / total);
    }
    return norm;
  }

  /**
   * Creates an add-1 smoothed Distribution from the given counter, ie adds one count
   * to every item, including unseen ones, and divides by the total count.
   * @param counter
   * @param numberOfKeys
   * @return a new add-1 smoothed Distribution
   */
  public static Distribution addOneSmoothedCounter(Counter counter, int numberOfKeys) {
    return lidstoneSmoothedCounter(counter, numberOfKeys, 1.0);
  }

  /**
   * Creates a smoothed Distribution using the Jeffreys-Perks law, ie adds one half
   * count to every item, including unseen ones, and divides by the total count.
   * @param counter
   * @param numberOfKeys
   * @return a new Jeffreys-Perks smoothed Distribution
   */
  public static Distribution jeffreysPerksSmoothedCounter(Counter counter, int numberOfKeys) {
    return lidstoneSmoothedCounter(counter, numberOfKeys, 0.5);
  }

  /**
   * Creates a smoothed Distribution using Lidstone's law, ie adds lambda (typically
   * between 0 and 1) to every item, including unseen ones, and divides by the total count.
   * @param counter
   * @param numberOfKeys
   * @param lambda
   * @return a new Lidstone smoothed Distribution
   */
  public static Distribution lidstoneSmoothedCounter(Counter counter, int numberOfKeys, double lambda) {
    Distribution norm = new Distribution();
    norm.counter = new Counter();
    double total = counter.totalCount() + (lambda * (double)numberOfKeys);
    double reservedMass = ((double)numberOfKeys - counter.size()) * lambda / total;
    norm.numberOfKeys = numberOfKeys;
    norm.reservedMass = reservedMass;
    for (Iterator iter = counter.keySet().iterator(); iter.hasNext(); ) {
      Object key = iter.next();
      norm.counter.setCount(key, (counter.getCount(key) + lambda) / total);
    }
    return norm;
  }

  /**
   * Creates a Good-Turing smoothed Distribution from the given counter.
   * @param counter
   * @param numberOfKeys
   * @return a new Good-Turing smoothed Distribution.
   */
  public static Distribution goodTuringSmoothedCounter(Counter counter, int numberOfKeys) {

    // gather count-counts
    int[] countCounts = new int[11];
    for (int i=0; i<=10; i++) countCounts[i] = 0;
    for (Iterator iter = counter.keySet().iterator(); iter.hasNext(); ) {
      Object key = iter.next();
      int count = (int)Math.round(counter.getCount(key));
      if (count <= 10)
        countCounts[count]++;
    }

    // if count-counts are unreliable, we shouldn't be using G-T
    // revert to J-P
    for (int i=1; i<=10; i++) {
      if (countCounts[i] < 3)
        return jeffreysPerksSmoothedCounter(counter, numberOfKeys);
    }

    double observedMass = counter.totalCount();
    double reservedMass = countCounts[1] / observedMass;

    // calculate and cache adjusted frequencies
    // also adjusting total mass of observed items
    double[] adjustedFreq = new double[10];
    for (int freq=1; freq<10; freq++) {
      adjustedFreq[freq] = (double)(freq+1) * (double)countCounts[freq+1] /
              (double)countCounts[freq];
      observedMass -= ((double)freq - adjustedFreq[freq]) * countCounts[freq];
    }

    double normFactor = (1.0-reservedMass) / observedMass;

    Distribution norm = new Distribution();
    norm.counter = new Counter();
    norm.numberOfKeys = numberOfKeys;
    norm.reservedMass = reservedMass;

    // fill in the new Distribution, renormalizing as we go
    for (Iterator iter = counter.keySet().iterator(); iter.hasNext(); ) {
      Object key = iter.next();
      int origFreq = (int)Math.round(counter.getCount(key));
      if (origFreq < 10) {
        norm.counter.setCount(key, adjustedFreq[origFreq] * normFactor);
      } else {
        norm.counter.setCount(key, (double)origFreq * normFactor);
      }
    }

//    System.out.println("totalCount: " + norm.totalCount());
    return norm;
  }

  /**
   * Returns a Distribution that uses prior as a Dirichlet prior
   * weighted by weight.  Essentially adds "pseudo-counts" for each Object
   * in prior proportional to that Object's mass in prior times weight,
   * then normalizes.
   *
   * WARNING: If unseen item is encountered in c, total may not be 1.
   *
   * @param c
   * @param prior
   * @param weight average "pseudo-count" of Objects in prior
   * @return new Distribution
   */
  public static Distribution normalizedCounterWithDirichletPrior(Counter c, Distribution prior, double weight) {

    double totalWeight = c.totalCount() + weight;
    double reservedMass = prior.reservedMass * weight / totalWeight;

    Distribution norm = new Distribution();
    norm.counter = Counters.linearCombination(c, 1 / totalWeight, prior.counter, weight / totalWeight);
    norm.numberOfKeys = prior.numberOfKeys;
    norm.reservedMass = reservedMass;
 //   System.out.println("totalCount: " + norm.totalCount());
    return norm;
  }


  /** Maps a counter representing the linear weights of a multiclass
   *  logistic regression model to the probabilities of each class.
   */
  public static Distribution distributionFromLogisticCounter(Counter cntr) {
    double expSum = 0.0;
    int numKeys = 0;
    for (Iterator iter = cntr.keySet().iterator(); iter.hasNext(); ) {
      Object key = iter.next();
      expSum += Math.exp(cntr.getCount(key));
      numKeys++;
    }
    Distribution probs = new Distribution();
    probs.counter = new Counter();
    probs.reservedMass = 0.0;
    probs.numberOfKeys = numKeys;
    for (Iterator iter = cntr.keySet().iterator(); iter.hasNext(); ) {
      Object key = iter.next();
      probs.counter.setCount(key, Math.exp(cntr.getCount(key))/expSum);
    }
    return probs;
  }


  /**
   * Returns the normalized count of the given object.
   * @param key
   * @return the normalized count of the object
   */
  public double probabilityOf(Object key) {
    if (counter.containsKey(key)) {
      return counter.getCount(key);
    } else if ((numberOfKeys - counter.keySet().size()) <= 0)
      return 0.0;
    else return (reservedMass / (numberOfKeys - counter.keySet().size()));
  }

  public Object argmax() {
    return counter.argmax();
  }

  private double totalCount() {
    return counter.totalCount() + reservedMass;
  }

  public boolean equals(Object o) {
    if (this == o) return true;
    if (!(o instanceof Distribution)) return false;

    final Distribution distribution = (Distribution) o;

    if (numberOfKeys != distribution.numberOfKeys) return false;
    if (reservedMass != distribution.reservedMass) return false;
    if (!counter.equals(distribution.counter)) return false;

    return true;
  }

  public int hashCode() {
    int result;
    long temp;
    result = numberOfKeys;
    temp = Double.doubleToLongBits(reservedMass);
    result = 29 * result + (int) (temp ^ (temp >>> 32));
    result = 29 * result + counter.hashCode();
    return result;
  }

  // can't instantiate this class; use static methods instead
  private Distribution() {
  }

  /**
   * For internal testing purposes only.
   */
  public static void main(String[] args) {
    Counter c = new Counter();

    final double p = 1000;

    // fill counter with roughly Zipfian distribution
    for (int rank=1; rank<2000; rank++) {
      c.setCount(new Integer(rank), Math.round(p/(double)rank));
    }

    Distribution n = normalizedCounter(c);
    Distribution add1 = addOneSmoothedCounter(c, 4000);
    Distribution jp = jeffreysPerksSmoothedCounter(c, 4000);
    Distribution gt = goodTuringSmoothedCounter(c, 4000);

    System.out.println("Freq  Norm                Add1                  J-P                  G-T");
    for (int i=1; i<5; i++) {
      System.out.print(Math.round(p/(double)i));
      System.out.print("   ");
      System.out.print(n.probabilityOf(new Integer(i)));
      System.out.print("  ");
      System.out.print(add1.probabilityOf(new Integer(i)));
      System.out.print("  ");
      System.out.print(jp.probabilityOf(new Integer(i)));
      System.out.print("  ");
      System.out.print(gt.probabilityOf(new Integer(i)));
      System.out.println();
    }
    System.out.println();
    System.out.println("--------------------------------------------");
    System.out.println();
    System.out.print(0);
    System.out.print("   ");
    System.out.print(n.probabilityOf(new Integer(0)));
    System.out.print("  ");
    System.out.print(add1.probabilityOf(new Integer(0)));
    System.out.print("  ");
    System.out.print(jp.probabilityOf(new Integer(0)));
    System.out.print("  ");
    System.out.print(gt.probabilityOf(new Integer(0)));
    System.out.println();

    System.out.println("Totals:");

    System.out.print(n.totalCount());
    System.out.print("  ");
    System.out.print(add1.totalCount());
    System.out.print("  ");
    System.out.print(jp.totalCount());
    System.out.print("  ");
    System.out.print(gt.totalCount());
    System.out.println();
  }

}
