package edu.stanford.nlp.util;

import edu.stanford.nlp.util.Filter;

import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;

/**
 * Iterator that suppresses items in another iterator based on a filter function.
 *
 * @author Dan Klein (klein@cs.stanford.edu)
 */
public class FilteredIterator implements Iterator {
  Iterator iterator = null;
  Filter filter = null;
  Object current = null;
  boolean hasCurrent = false;
  Object currentCandidate() {
    return current;
  }
  void advanceCandidate() {
    if (! iterator.hasNext()) {
      hasCurrent = false;
      current = null;
      return;
    }
    hasCurrent = true;
    current = iterator.next();
  }
  boolean hasCurrentCandidate() {
    return hasCurrent;
  }
  boolean currentCandidateIsAcceptable() {
    return filter.accept(currentCandidate());
  }
  void skipUnacceptableCandidates() {
    while (hasCurrentCandidate() && ! currentCandidateIsAcceptable()) {
      advanceCandidate();
    }
  }
  public boolean hasNext() {
    return hasCurrentCandidate();
  }
  public Object next() {
    Object result = currentCandidate();
    advanceCandidate();
    skipUnacceptableCandidates();
    return result;
  }
  public void remove() {
    throw new UnsupportedOperationException();
  }
  public FilteredIterator(Iterator iterator, Filter filter) {
    this.iterator = iterator;
    this.filter = filter;
    advanceCandidate();
    skipUnacceptableCandidates();
  }
  public static void main(String[] args) {
    Collection c = Arrays.asList(new String[] {"a", "aa", "b", "bb", "cc"});
    Iterator i = new FilteredIterator(c.iterator(),
                                      new Filter() {
                                        public boolean accept(Object o) {
                                          return o.toString().length() == 1;
                                      }
                                      });
    while (i.hasNext()) {
      System.out.println("Accepted: "+i.next());
    }
  }
}
