package edu.stanford.nlp.util;

import edu.stanford.nlp.util.StringUtils;

/**
 * Similarity metric based on longest-common-substrings
 * between two strings. Based on Edit Distance code taken from 
 * http://www.merriampark.com/ld.htm and the algorithm described
 * in http://www2.toki.or.id/book/AlgDesignManual/BOOK/BOOK5/NODE208.HTM
 * This class can be used as a Similarity class, or as a static class
 * for computing arbitrary LCS.
 */
public class LongestCommonSubstring
{
    private static boolean verbose=false;

    /**
     * Returns the minimum of three values.
     */
    private static int Maximum (int a, int b, int c)
    {
        int ma;

        ma = a;
        if (b > ma) ma = b;
        if (c > ma) ma = c;
        return ma;

    }

    /**
     * Computes the longest common substring of s and t.
     * The longest common substring of a and b is the longest run of 
     * characters that appear in order inside both a and b. Both a and b
     * may have other extraneous characters along the way. This is like
     * edit distance but with no substitution and a higher number means
     * more similar. For example, the LCS of "abcD" and "aXbc" is 3 (abc).
     */
    public static int LCS(String s, String t) 
    {
        int d[][]; // matrix
        int n; // length of s
        int m; // length of t
        int i; // iterates through s
        int j; // iterates through t
        char s_i; // ith character of s
        char t_j; // jth character of t
        int cost; // cost

        // Step 1

        n = s.length ();
        m = t.length ();
        if (n == 0) return 0;
        if (m == 0) return 0;
        d = new int[n+1][m+1];

        // Step 2

        for (i = 0; i <= n; i++) d[i][0] = 0;
        for (j = 0; j <= m; j++) d[0][j] = 0;

        // Step 3
        for (i = 1; i <= n; i++) 
        {
            s_i = s.charAt (i - 1);

            // Step 4
            for (j = 1; j <= m; j++) 
            {
                t_j = t.charAt (j - 1);

                // Step 5 
                // js: if the chars match, you can get an extra point
                // otherwise you have to skip an insertion or deletion (no subs)
                if (s_i == t_j) d[i][j] = Maximum (d[i-1][j], d[i][j-1], d[i-1][j-1]+1);
                else d[i][j] = Math.max (d[i-1][j], d[i][j-1]);
            }
        }
        if(verbose)
        {
            // num chars needed to display longest num
            int numChars=(int)Math.ceil(Math.log(d[n][m])/Math.log(10)); 
            for(i=0;i<numChars+3;i++) System.err.print(' ');
            for(j=0;j<m;j++) 
                System.err.print(StringUtils.leftPad(""+t.charAt(j),numChars)+" "); 
            System.err.println();
            for(i=0;i<=n;i++)
            {
                System.err.print((i==0?' ':s.charAt(i-1))+" ");
                for(j=0;j<=m;j++)
                    System.err.print(StringUtils.leftPad(""+d[i][j],numChars)+" ");
                System.err.println();
            }
        }

        // Step 7
        return d[n][m];
    }

    /** For internal debugging purposes only. */
    public static void main(String[] args)
    {
        LongestCommonSubstring lcs=new LongestCommonSubstring();
        lcs.verbose=true;
        System.err.println(lcs.LCS("Jo3seph Smarr!","Joseph R Smarr"));
    }
}

