package edu.stanford.nlp.util;

import java.util.*;
import java.util.regex.*;
import java.io.*;
import java.net.*;

/**
 * Stringtils is a class for random String things.
 *
 * @author Dan Klein
 * @author Christopher Manning
 * @author Tim Grow (grow@stanford.edu)
 * @version 2003/02/03
 */
public class StringUtils {

    /**
     * Don't let anyone instantiate this class.
     */
    private StringUtils() {
    }


    /**
     * Say whether this regular expression can be found inside
     * this String.  This method provides one of the two "missing"
     * convenience methods for regular expressions in the String class
     * in JDK1.4.  This is the one you'll want to use all the time if
     * you're used to Perl.  What were they smoking?
     *
     * @param str   String to search for match in
     * @param regex String to compile as the regular expression
     * @return Whether the regex can be found in str
     */
    public static boolean find(String str, String regex) {
        return Pattern.compile(regex).matcher(str).find();
    }


    /**
     * Say whether this regular expression can be found at the beginning of
     * this String.  This method provides one of the two "missing"
     * convenience methods for regular expressions in the String class
     * in JDK1.4.
     *
     * @param str   String to search for match at start of
     * @param regex String to compile as the regular expression
     * @return Whether the regex can be found at the start of str
     */
    public static boolean lookingAt(String str, String regex) {
        return Pattern.compile(regex).matcher(str).lookingAt();
    }


    /**
     * Say whether this regular expression matches
     * this String.  This method is the same as the String.matches() method,
     * and is included just to give a call that is parallel to the other
     * static regex methods in this class.
     *
     * @param str   String to search for match at start of
     * @param regex String to compile as the regular expression
     * @return Whether the regex matches the whole of this str
     */
    public static boolean matches(String str, String regex) {
        return Pattern.compile(regex).matcher(str).matches();
    }


    private static final int SLURPBUFFSIZE = 16000;

    /**
     * Returns all the text in the given File.
     */
    public static String slurpFile(File file) throws IOException {
	char[] chars = new char[SLURPBUFFSIZE];
        Reader r = new FileReader(file);
        StringBuffer buff = new StringBuffer();

	while (true) {
	    int amountRead = r.read(chars, 0, SLURPBUFFSIZE);
	    if (amountRead < 0) {
		break;
	    }
	    buff.append(chars, 0, amountRead);
	}
        r.close();
        return buff.toString();
    }


    /**
     * Returns all the text in the given File.
     */
    public static String slurpFile(String filename) {
        StringBuffer buff = new StringBuffer();
	char[] chars = new char[SLURPBUFFSIZE];
	Reader r = null;
	try {
	    r = new FileReader(filename);

	    while (true) {
		int amountRead = r.read(chars, 0, SLURPBUFFSIZE);
		if (amountRead < 0) {
		    break;
		}
		buff.append(chars, 0, amountRead);
	    }
	} catch (IOException ioe) {
	} finally {
	    try {
		if (r != null) {
		    r.close();
		}
	    } catch (IOException ioe) {
	    }
	}
        return buff.toString();
    }


    /**
     * Returns all the text at the given URL.
     */
    public static String slurpURL(URL u) throws IOException {
        String lineSeparator = System.getProperty("line.separator");
        URLConnection uc = u.openConnection();
        InputStream is = uc.getInputStream();
        BufferedReader br = new BufferedReader(new InputStreamReader(is));
        String temp;
        StringBuffer buff = new StringBuffer(16000);  // make biggish
        while ((temp = br.readLine()) != null) {
            buff.append(temp);
            buff.append(lineSeparator);
        }
        br.close();
        return buff.toString();
    }

    /**
     * Joins each elem in the List with the given glue. For example, given a
     * list
     * of Integers, you can create a comma-separated list by calling
     * <tt>join(numbers, ", ")</tt>.
     */
    public static String join(List l, String glue) {
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < l.size(); i++) {
            if (i > 0)
                sb.append(glue);
            sb.append(l.get(i).toString());
        }
        return sb.toString();
    }

    /**
     * Joins each elem in the array with the given glue. For example, given a list
     * of ints, you can create a comma-separated list by calling
     * <tt>join(numbers, ", ")</tt>.
     */
    public static String join(Object[] elements, String glue) {
        return (join(Arrays.asList(elements), glue));
    }

    /**
     * Joins elems with a space.
     */
    public static String join(List l) {
        return join(l, " ");
    }

    /**
     * Joins elems with a space.
     */
    public static String join(Object[] elements) {
        return (join(elements, " "));
    }


    /**
     * Splits on whitespace (\\s+).
     */
    public static List split(String s) {
        return (split(s, "\\s+"));
    }

    /**
     * Splits the given string using the given regex as delimiters.
     * This method is the same as the String.split() method (except it throws
     * the results in a List),
     * and is included just to give a call that is parallel to the other
     * static regex methods in this class.
     *
     * @param str   String to split up
     * @param regex String to compile as the regular expression
     * @return List of Strings resulting from splitting on the regex
     */
    public static List split(String str, String regex) {
        return (Arrays.asList(str.split(regex)));
    }


    /**
     * Return a String of length a minimum of totalChars characters by
     * padding the input String str with spaces.  If str is already longer
     * than totalChars, it is returned unchanged.
     */
    public static String pad(String str, int totalChars) {
        int slen = str.length();
        if (slen > totalChars) {
            return str;
        } else {
            StringBuffer sb = new StringBuffer(str);
            for (int i = 0; i < totalChars - slen; i++) {
                sb.append(" ");
            }
            return sb.toString();
        }
    }

    /**
     * Pads the toString value of the given Object.
     */
    public static String pad(Object obj, int totalChars) {
        return pad(obj.toString(), totalChars);
    }


    /**
     * Pad or trim so as to produce a string of exactly a certain length.
     *
     * @param inStr The String to be padded or truncated
     * @param num   The desired length
     */
    public static String exactlyN(String inStr, int num) {
        int leng = inStr.length();
        if (leng < num) {
            StringBuffer sb = new StringBuffer(inStr);
            for (int i = 0; i < num - leng; i++) {
                sb.append(" ");
            }
            return sb.toString();
        } else if (leng > num) {
            return inStr.substring(0, num);
        } else {
            return inStr;
        }
    }

    /**
     * Pad or trim the toString value of the given Object.
     */
    public static String exactlyN(Object obj, int totalChars) {
        return exactlyN(obj.toString(), totalChars);
    }


    /**
     * Pads the given String to the left with spaces to ensure that it's
     * at least totalChars long.
     */
    public static String leftPad(String str, int totalChars) {
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < totalChars - str.length(); i++) {
            sb.append(" ");
        }
        sb.append(str);
        return sb.toString();
    }


    public static String leftPad(Object obj, int totalChars) {
        return leftPad(obj.toString(), totalChars);
    }

    public static String leftPad(int i, int totalChars) {
        return leftPad(new Integer(i), totalChars);
    }

    public static String leftPad(double d, int totalChars) {
        return leftPad(new Double(d), totalChars);
    }

    /**
     * Returns s if it's at most maxWidth chars, otherwise chops right side to fit.
     */
    public static String trim(String s, int maxWidth) {
        if (s.length() <= maxWidth) return (s);
        return (s.substring(0, maxWidth));
    }

    public static String trim(Object obj, int maxWidth) {
        return trim(obj.toString(), maxWidth);
    }

    /**
     * Returns a "clean" version of the given filename in which spaces have
     * been converted to dashes and all non-alphaneumeric chars are underscores.
     */
    public static String fileNameClean(String s) {
        char[] chars = s.toCharArray();
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < chars.length; i++) {
            char c = chars[i];
            if ((c >= 'A' && c <= 'Z') ||
                    (c >= 'a' && c <= 'z') ||
                    (c >= '0' && c <= '9') ||
                    (c == '_'))
                sb.append(c);
            else {
                if (c == ' ' ||
                        c == '-')
                    sb.append('_');
                else
                    sb.append("x" + (int) c + "x");
            }
        }
        return sb.toString();
    }

    /**
     * Returns the index of the <i>n</i>th occurrence of ch in s, or -1
     * if there are less than n occurrences of ch.
     */
    public static int nthIndex(String s, char ch, int n) {
        int index = 0;
        for (int i = 0; i < n; i++) {
            // if we're already at the end of the string,
            // and we need to find another ch, return -1
            if (index == s.length() - 1) return -1;
            index = s.indexOf(ch, index + 1);
            if (index == -1) return (-1);
        }
        return index;
    }


    /**
     * This returns a string from decimal digit smallestDigit to decimal digit
     * biggest digit. Smallest digit is labeled 1, and the limits are
     * inclusive.
     */
    public static String paddedIntToString(int n, int smallestDigit,
                                           int biggestDigit) {
        int numDigits = biggestDigit - smallestDigit + 1;
        char[] result = new char[numDigits];
        for (int j = 1; j < smallestDigit; j++) {
            n = n / 10;
        }
        for (int j = numDigits - 1; j >= 0; j--) {
            result[j] = Character.forDigit(n % 10, 10);
            n = n / 10;
        }
        return new String(result);
    }

    /**
     * Parses command line arguments into a Map. Arguments of the form
     * <p/>
     * -flag1 arg1a arg1b ... arg1m -flag2 -flag3 arg3a ... arg3n
     * <p/>
     * will be parsed so that the flag is a key in the Map (including
     * the hyphen) and its value will be a {@link String[] } containing
     * the optional arguments (if present).  The non-flag values not
     * captured as flag arguments are collected into a String[] array
     * and returned as the value of <code>null</code> in the Map.  In
     * this invocation, flags cannot take arguments, so all the {@link
     * String} array values other than the value for <code>null</code>
     * will be zero-length.
     *
     * @param args
     * @return a {@link Map} of flag names to flag argument {@link
     *         String[]} arrays.
     */
    public static Map argsToMap(String[] args) {
        return argsToMap(args, new HashMap());
    }

    /**
     * Parses command line arguments into a Map. Arguments of the form
     * <p/>
     * -flag1 arg1a arg1b ... arg1m -flag2 -flag3 arg3a ... arg3n
     * <p/>
     * will be parsed so that the flag is a key in the Map (including
     * the hyphen) and its value will be a {@link String[] } containing
     * the optional arguments (if present).  The non-flag values not
     * captured as flag arguments are collected into a String[] array
     * and returned as the value of <code>null</code> in the Map.  In
     * this invocation, the maximum number of arguments for each flag
     * can be specified as an {@link Integer} value of the appropriate
     * flag key in the <code>flagsToNumArgs</code> {@link Map}
     * argument. (By default, flags cannot take arguments.)
     * <p/>
     * Example of usage:
     * <p/>
     * <code>
     * Map flagsToNumArgs = new HashMap();
     * flagsToNumArgs.put("-x",new Integer(2));
     * flagsToNumArgs.put("-d",new Integer(1));
     * Map result = parseCommandLineArguments(args,flagsToNumArgs);
     * </code>
     *
     * @param args           the argument array to be parsed
     * @param flagsToNumArgs a {@link Map} of flag names to {@link
     *                       Integer} values specifying the maximum number of allowed
     *                       arguments for that flag (default 0).
     * @return a {@link Map} of flag names to flag argument {@link
     *         String[]} arrays.
     */
    public static Map argsToMap(String[] args, Map flagsToNumArgs) {
        Map result = new HashMap();
        List remainingArgs = new ArrayList();
        String key, value;
        for (int i = 0; i < args.length; i++) {
            key = args[i];
            if (key.charAt(0) == '-') { // found a flag
                Integer maxFlagArgs = (Integer) flagsToNumArgs.get(key);
                int max = maxFlagArgs == null ? 0 : maxFlagArgs.intValue();
                List flagArgs = new ArrayList();
                for (int j = 0; j < max && i + 1 < args.length && args[i + 1].charAt(0) != '-'; i++, j++) {
                    flagArgs.add(args[i + 1]);
                }
                result.put(key, flagArgs.toArray(new String[]{}));
            } else {
                remainingArgs.add(args[i]);
            }
        }
        result.put(null, remainingArgs.toArray(new String[]{}));
        return result;
    }

    private static final String PROP = "prop";

    public static Properties argsToProperties(String[] args) {
        return argsToProperties(args, new HashMap());
    }

    /**
     * Analagous to {@link #argsToMap}.  However, there are several key differences between this method and {@link #argsToMap}:
     * <ul>
     * <li> Hyphens are stripped from flag names </li>
     * <li> Since Properties objects are String to String mappings, the default number of arguments to a flag is
     * assumed to be 1 and not 0. </li>
     * <li> Furthermore, the list of arguments not bound to a flag is mapped to the "" property, not null </li>
     * <li> The special flag "-prop" will load the property file specified by it's argument. </li>
     * <li> The value for flags without arguments is set to "true" </li>
     */
    public static Properties argsToProperties(String[] args, Map flagsToNumArgs) {
        Properties result = new Properties();
        List remainingArgs = new ArrayList();
        String key;
        for (int i = 0; i < args.length; i++) {
            key = args[i];
            if (key.charAt(0) == '-') { // found a flag
                key = key.substring(1); // strip off the hyphen

                Integer maxFlagArgs = (Integer) flagsToNumArgs.get(key);
                int max = maxFlagArgs == null ? 1 : maxFlagArgs.intValue();
                List flagArgs = new ArrayList();
                for (int j = 0; j < max && i + 1 < args.length && args[i + 1].charAt(0) != '-'; i++, j++) {
                    flagArgs.add(args[i + 1]);
                }
                if (flagArgs.size() == 0)
                    result.setProperty(key, "true");
                else {
                    result.setProperty(key, join(flagArgs, " "));
                    if (key.equalsIgnoreCase(PROP)) {
                        try {
                            result.load(new BufferedInputStream(new FileInputStream(result.getProperty(PROP))));
                        } catch (IOException e) {
                            e.printStackTrace();
                        }
                    }
                }
            } else {
                remainingArgs.add(args[i]);
            }
        }
        result.setProperty("", join(remainingArgs, " "));
        return result;
    }


    /**
     * Prints to a file.  If the file already exists, appends if
     * <code>append=true</code>, and overwrites if <code>append=false</code>
     */
    public static void printToFile(String filename, String message, boolean append) {
        FileWriter fw = null;
        PrintWriter pw = null;
        try {
            fw = new FileWriter(filename, append);
            pw = new PrintWriter(fw);
            pw.print(message);
        } catch (Exception e) {
            System.out.println("Exception: in printToFile " + filename + " " +
                    message);
            e.printStackTrace();
        } finally {
            if (pw != null) pw.close();
        }
    }


    /**
     * Prints to a file.  If the file does not exist, rewrites the file;
     * does not append.
     */
    public static void printToFile(String filename, String message) {
        printToFile(filename, message, false);
    }

  /**
   * THIS IS HIGHLY SUPERIOR TO THE OVERLY COMPLEXIFIED CODE THAT COMES BEFORE IT.
   * Parses command line arguments into a Map. Arguments of the form
   * -flag1 arg1 -flag2 -flag3 arg3
   * will be parsed so that the flag is a key in the Map (including the hyphen) and the
   * optional argument will be its value (if present).
   * @param args
   * @return
   */
  public static Map parseCommandLineArguments(String[] args) {
    Map result = new HashMap();
    String key, value;
    for (int i = 0; i < args.length; i++) {
      key = args[i];
      if (key.charAt(0) == '-') {
        if (i + 1 < args.length) {
          value = args[i + 1];
          if (value.charAt(0) != '-') {
            result.put(key, value);
            i++;
          } else {
            result.put(key, null);
          }
        } else {
          result.put(key, null);
        }
      }
    }
    return result;
  }


}
